import contextlib
import datetime
import threading
from collections.abc import Iterable
from datetime import timedelta
from typing import List, cast

import pytest
from pytest_django.fixtures import SettingsWrapper
from pytest_mock import MockerFixture

import anyio

from django.core.cache import caches
from django.core.cache.backends.base import DEFAULT_TIMEOUT
from django.test import override_settings

from django_valkey.async_cache.cache import AsyncValkeyCache
from django_valkey.async_cache.client import AsyncHerdClient
from django_valkey.serializers.json import JSONSerializer
from django_valkey.serializers.msgpack import MSGPackSerializer


pytestmark = pytest.mark.anyio


@pytest.fixture
async def patch_itersize_setting() -> Iterable[None]:
    del caches["default"]
    with override_settings(DJANGO_VALKEY_SCAN_ITERSIZE=30):
        yield

    del caches["default"]


class TestAsyncDjangoValkeyCache:
    async def test_set_int(self, cache: AsyncValkeyCache):
        if isinstance(cache.client, AsyncHerdClient):
            pytest.skip("Herd client's set method works differently")
        await cache.aset("test_key", 1)
        result = await cache.aget("test_key")
        assert type(result) is int
        raw_client = await cache.client._get_client(write=False, client=None)
        assert await raw_client.get(":1:test_key") == b"1"

    async def test_set_float(self, cache: AsyncValkeyCache):
        if isinstance(cache.client, AsyncHerdClient):
            pytest.skip("Herd client's set method works differently")
        await cache.aset("test_key2", 1.1)
        result = await cache.aget("test_key2")
        assert type(result) is float
        raw_client = await cache.client._get_client(write=False, client=None)
        assert await raw_client.get(":1:test_key2") == b"1.1"

    async def test_setnx(self, cache: AsyncValkeyCache):
        await cache.delete("test_key_nx")
        res = await cache.get("test_key_nx")
        assert res is None

        res = await cache.set("test_key_nx", 1, nx=True)
        assert bool(res) is True

        res = await cache.set("test_key_nx", 2, nx=True)
        assert res is None
        res = await cache.get("test_key_nx")
        assert res == 1

        await cache.delete("test_key_nx")
        res = await cache.get("test_key_nx")
        assert res is None

    async def test_setnx_timeout(self, cache: AsyncValkeyCache):
        # test that timeout still works for nx=True
        res = await cache.aset("test_key_nx", 1, timeout=2, nx=True)
        assert res is True
        await anyio.sleep(3)
        res = await cache.aget("test_key_nx")
        assert res is None

        # test that timeout will not affect key, if it was there
        await cache.aset("test_key_nx", 1)
        res = await cache.aset("test_key_nx", 2, timeout=2, nx=True)
        assert res is None
        await anyio.sleep(3)
        res = await cache.aget("test_key_nx")
        assert res == 1

        await cache.adelete("test_key_nx")
        res = await cache.aget("test_key_nx")
        assert res is None

    async def test_unicode_keys(self, cache: AsyncValkeyCache):
        await cache.aset("ключ", "value")
        res = await cache.aget("ключ")
        assert res == "value"

    async def test_save_and_integer(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", 2)
        res = await cache.aget("test_key", "Foo")

        assert isinstance(res, int)
        assert res == 2

    async def test_save_string(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", "hello" * 1000)
        res = await cache.aget("test_key")

        assert isinstance(res, str)
        assert res == "hello" * 1000

        await cache.aset("test_key", "2")
        res = await cache.aget("test_key")

        assert isinstance(res, str)
        assert res == "2"

    async def test_save_unicode(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", "heló")
        res = await cache.aget("test_key")

        assert isinstance(res, str)
        assert res == "heló"

        async def test_save_dict(self, cache: AsyncValkeyCache):
            if isinstance(
                cache.client._serializer, (JSONSerializer, MSGPackSerializer)
            ):
                # JSONSerializer and MSGPackSerializer use the isoformat for
                # datetimes.
                now_dt: str | datetime.datetime = datetime.datetime.now().isoformat()
            else:
                now_dt = datetime.datetime.now()

            test_dict = {"id": 1, "date": now_dt, "name": "Foo"}

            await cache.aset("test_key", test_dict)
            res = await cache.aget("test_key")

            assert isinstance(res, dict)
            assert res["id"] == 1
            assert res["name"] == "Foo"
            assert res["date"] == now_dt

    async def test_save_float(self, cache: AsyncValkeyCache):
        float_val = 1.345620002

        await cache.aset("test_key", float_val)
        res = await cache.aget("test_key")

        assert isinstance(res, float)
        assert res == float_val

    async def test_timeout(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", 222, timeout=3)
        await anyio.sleep(4)

        res = await cache.aget("test_key")
        assert res is None

    async def test_timeout_0(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", 222, timeout=0)
        res = await cache.aget("test_key")
        assert res is None

    async def test_timeout_parameter_as_positional_argument(
        self, cache: AsyncValkeyCache
    ):
        await cache.aset("test_key", 222, -1)
        res = await cache.aget("test_key")
        assert res is None

        await cache.aset("test_key", 222, 1)
        res1 = await cache.aget("test_key")
        await anyio.sleep(2)
        res2 = await cache.aget("test_key")
        assert res1 == 222
        assert res2 is None

        # nx=True should not overwrite expire of key already in db
        await cache.aset("test_key", 222, None)
        await cache.aset("test_key", 222, -1, nx=True)
        res = await cache.aget("test_key")
        assert res == 222

    async def test_timeout_negative(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", 222, timeout=-1)
        res = await cache.aget("test_key")
        assert res is None

        await cache.aset("test_key", 222, timeout=None)
        await cache.aset("test_key", 222, timeout=-1)
        res = await cache.aget("test_key")
        assert res is None

        # nx=True should not overwrite expire of key already in db
        await cache.aset("test_key", 222, timeout=None)
        await cache.aset("test_key", 222, timeout=-1, nx=True)
        res = await cache.aget("test_key")
        assert res == 222

    async def test_timeout_tiny(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", 222, timeout=0.00001)
        res = await cache.aget("test_key")
        assert res in (None, 222)

    async def test_set_add(self, cache: AsyncValkeyCache):
        await cache.aset("add_key", "Initial value")
        res = await cache.aadd("add_key", "New value")
        assert res is None

        res = await cache.aget("add_key")
        assert res == "Initial value"
        res = await cache.aadd("other_key", "New value")
        assert res is True

    async def test_get_many(self, cache: AsyncValkeyCache):
        await cache.aset("a", 1)
        await cache.aset("b", 2)
        await cache.aset("c", 3)

        res = await cache.aget_many(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    async def test_mget(self, cache: AsyncValkeyCache):
        await cache.aset("a", 1)
        await cache.aset("b", 2)
        await cache.aset("c", 3)

        res = await cache.mget(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    async def test_get_many_unicode(self, cache: AsyncValkeyCache):
        await cache.aset("a", "1")
        await cache.aset("ب", "2")
        await cache.aset("c", "الف")

        res = await cache.aget_many(["a", "ب", "c"])
        assert res == {"a": "1", "ب": "2", "c": "الف"}

    async def test_mget_unicode(self, cache: AsyncValkeyCache):
        await cache.aset("a", "1")
        await cache.aset("ب", "2")
        await cache.aset("c", "الف")

        res = await cache.mget(["a", "ب", "c"])
        assert res == {"a": "1", "ب": "2", "c": "الف"}

    async def test_set_many(self, cache: AsyncValkeyCache):
        await cache.aset_many({"a": 1, "b": 2, "c": 3})
        res = await cache.aget_many(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    async def test_mset(self, cache: AsyncValkeyCache):
        await cache.amset({"a": 1, "b": 2, "c": 3})
        res = await cache.aget_many(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    async def test_set_call_empty_pipeline(
        self,
        cache: AsyncValkeyCache,
        mocker: MockerFixture,
        settings: SettingsWrapper,
    ):
        tmp = await cache.client.get_client(write=True)
        pipeline = await tmp.pipeline()
        key = "key"
        value = "value"

        mocked_set = mocker.patch.object(pipeline, "set", mocker.AsyncMock())

        await cache.aset(key, value, client=pipeline)

        if isinstance(cache.client, AsyncHerdClient):
            default_timeout = cache.client._backend.default_timeout
            herd_timeout = (default_timeout + settings.CACHE_HERD_TIMEOUT) * 1000
            herd_pack_value = cache.client._pack(value, default_timeout)
            mocked_set.assert_awaited_once_with(
                cache.client.make_key(key, version=None),
                cache.client.encode(herd_pack_value),
                nx=False,
                px=herd_timeout,
                xx=False,
            )
        else:
            mocked_set.assert_awaited_once_with(
                cache.client.make_key(key, version=None),
                cache.client.encode(value),
                nx=False,
                px=cache.client._backend.default_timeout * 1000,
                xx=False,
            )

    async def test_delete(self, cache: AsyncValkeyCache):
        await cache.aset_many({"a": 1, "b": 2, "c": 3})
        res = await cache.adelete("a")
        assert res is True

        res = await cache.aget_many(["a", "b", "c"])
        assert res == {"b": 2, "c": 3}

        res = await cache.adelete("a")
        assert res is False

    async def test_delete_return_value_type(self, cache: AsyncValkeyCache):
        await cache.aset("a", 1)
        res = await cache.adelete("a")
        assert isinstance(res, bool)
        assert res is True
        res = await cache.adelete("b")
        assert isinstance(res, bool)
        assert res is False

    async def test_delete_many(self, cache: AsyncValkeyCache):
        await cache.aset_many({"a": 1, "b": 2, "c": 3})
        res = await cache.adelete_many(["a", "b"])
        assert res == 2

        res = await cache.aget_many(["a", "b", "c"])
        assert res == {"c": 3}

        res = await cache.adelete_many(["a", "b"])
        assert res == 0

    async def test_delete_many_generator(self, cache: AsyncValkeyCache):
        await cache.aset_many({"a": 1, "b": 2, "c": 3})
        res = await cache.adelete_many(key for key in ["a", "b"])
        assert res == 2

        res = await cache.aget_many(["a", "b", "c"])
        assert res == {"c": 3}

        res = await cache.adelete_many((key for key in ("a", "b")))
        assert res == 0

    async def test_delete_many_empty_generator(self, cache: AsyncValkeyCache):
        res = await cache.adelete_many(key for key in cast(List[str], []))
        assert res == 0

    async def test_incr(self, cache: AsyncValkeyCache):
        if isinstance(cache.client, AsyncHerdClient):
            pytest.skip("HerdClient doesn't support incr")

        await cache.aset("num", 1)

        await cache.aincr("num")
        res = await cache.aget("num")
        assert res == 2

        await cache.aincr("num", 10)
        res = await cache.aget("num")
        assert res == 12

        # max 64-bit signed int
        await cache.aset("num", 9223372036854775807)

        await cache.aincr("num")
        res = await cache.aget("num")
        assert res == 9223372036854775808

        await cache.aincr("num", 2)
        res = await cache.aget("num")
        assert res == 9223372036854775810

        await cache.aset("num", 3)

        await cache.aincr("num", 2)
        res = await cache.aget("num")
        assert res == 5

    async def test_incr_no_timeout(self, cache: AsyncValkeyCache):
        if isinstance(cache.client, AsyncHerdClient):
            pytest.skip("HerdClient doesn't support incr")

        await cache.aset("num", 1, timeout=None)

        await cache.aincr("num")
        res = await cache.aget("num")
        assert res == 2

        await cache.aincr("num", 10)
        res = await cache.aget("num")
        assert res == 12

        # max 64-bit signed int
        await cache.aset("num", 9223372036854775807, timeout=None)

        await cache.aincr("num")
        res = await cache.aget("num")
        assert res == 9223372036854775808

        await cache.aincr("num", 2)
        res = await cache.aget("num")
        assert res == 9223372036854775810

        await cache.aset("num", 3, timeout=None)

        await cache.aincr("num", 2)
        res = await cache.aget("num")
        assert res == 5

    async def test_incr_error(self, cache: AsyncValkeyCache):
        if isinstance(cache.client, AsyncHerdClient):
            pytest.skip("HerdClient doesn't support incr")

        with pytest.raises(ValueError):
            # key does not exist
            await cache.aincr("numnum")

    async def test_incr_ignore_check(self, cache: AsyncValkeyCache):
        if isinstance(cache.client, AsyncHerdClient):
            pytest.skip("HerdClient doesn't support incr")

        # key exists check will be skipped and the value will be incremented by
        # '1' which is the default delta
        await cache.aincr("num", ignore_key_check=True)
        res = await cache.aget("num")
        assert res == 1
        await cache.adelete("num")

        # since key doesn't exist it is set to the delta value, 10 in this case
        await cache.aincr("num", 10, ignore_key_check=True)
        res = await cache.aget("num")
        assert res == 10
        await cache.adelete("num")

        # following are just regression checks to make sure it still works as
        # expected with incr max 64-bit signed int
        await cache.aset("num", 9223372036854775807)

        await cache.aincr("num", ignore_key_check=True)
        res = await cache.aget("num")
        assert res == 9223372036854775808

        await cache.aincr("num", 2, ignore_key_check=True)
        res = await cache.aget("num")
        assert res == 9223372036854775810

        await cache.aset("num", 3)

        await cache.aincr("num", 2, ignore_key_check=True)
        res = await cache.aget("num")
        assert res == 5

    async def test_get_set_bool(self, cache: AsyncValkeyCache):
        await cache.aset("bool", True)
        res = await cache.aget("bool")

        assert isinstance(res, bool)
        assert res is True

        await cache.aset("bool", False)
        res = await cache.aget("bool")

        assert isinstance(res, bool)
        assert res is False

    async def test_decr(self, cache: AsyncValkeyCache):
        if isinstance(cache.client, AsyncHerdClient):
            pytest.skip("HerdClient doesn't support decr")

        await cache.aset("num", 20)

        await cache.adecr("num")
        res = await cache.aget("num")
        assert res == 19

        await cache.adecr("num", 20)
        res = await cache.aget("num")
        assert res == -1

        await cache.adecr("num", 2)
        res = await cache.aget("num")
        assert res == -3

        await cache.aset("num", 20)

        await cache.adecr("num")
        res = await cache.aget("num")
        assert res == 19

        # max 64-bit signed int + 1
        await cache.aset("num", 9223372036854775808)

        await cache.adecr("num")
        res = await cache.aget("num")
        assert res == 9223372036854775807

        await cache.adecr("num", 2)
        res = await cache.aget("num")
        assert res == 9223372036854775805

    async def test_version(self, cache: AsyncValkeyCache):
        await cache.aset("keytest", 2, version=2)
        res = await cache.aget("keytest")
        assert res is None

        res = await cache.aget("keytest", version=2)
        assert res == 2

    async def test_incr_version(self, cache: AsyncValkeyCache):
        await cache.aset("keytest", 2)
        await cache.aincr_version("keytest")

        res = await cache.aget("keytest")
        assert res is None

        res = await cache.aget("keytest", version=2)
        assert res == 2

    async def test_ttl_incr_version_no_timeout(self, cache: AsyncValkeyCache):
        await cache.aset("my_key", "hello world!", timeout=None)

        await cache.aincr_version("my_key")

        my_value = await cache.aget("my_key", version=2)

        assert my_value == "hello world!"

    async def test_decr_version(self, cache: AsyncValkeyCache):
        await cache.aset("keytest", 2, version=3)
        res = await cache.aget("keytest", version=3)
        assert res == 2

        await cache.decr_version("keytest", version=3)

        res = await cache.aget("keytest", version=3)
        assert res is None

        res = await cache.aget("keytest", version=2)
        assert res == 2

    async def test_ttl_decr_version_no_timeout(self, cache: AsyncValkeyCache):
        await cache.set("my_key", "hello world!", timeout=None, version=3)

        await cache.adecr_version("my_key", version=3)

        my_value = await cache.get("my_key", version=2)

        assert my_value == "hello world!"

    async def test_delete_pattern(self, cache: AsyncValkeyCache):
        for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
            await cache.aset(key, "foo")

        res = await cache.adelete_pattern("*foo-a*")
        assert bool(res) is True

        keys = await cache.akeys("foo*")
        assert set(keys) == {"foo-bb", "foo-bc"}

        res = await cache.adelete_pattern("*foo-a*")
        assert bool(res) is False

    async def test_delete_pattern_with_custom_count(
        self, cache: AsyncValkeyCache, mocker
    ):
        client_mock = mocker.patch(
            "django_valkey.async_cache.cache.AsyncValkeyCache.client",
            new_callable=mocker.AsyncMock,
        )

        for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
            await cache.aset(key, "foo")

        await cache.adelete_pattern("*foo-a*", itersize=2)

        client_mock.adelete_pattern.assert_awaited_once_with("*foo-a*", itersize=2)

    async def test_delete_pattern_with_settings_default_scan_count(
        self,
        patch_itersize_setting,
        cache: AsyncValkeyCache,
        settings: SettingsWrapper,
        mocker,
    ):
        client_mock = mocker.patch(
            "django_valkey.async_cache.cache.AsyncValkeyCache.client",
            new_callable=mocker.AsyncMock,
        )

        for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
            await cache.aset(key, "foo")
        expected_count = settings.DJANGO_VALKEY_SCAN_ITERSIZE

        await cache.adelete_pattern("*foo-a*")

        client_mock.adelete_pattern.assert_awaited_once_with(
            "*foo-a*", itersize=expected_count
        )

    async def test_close(self, cache: AsyncValkeyCache, settings: SettingsWrapper):
        settings.DJANGO_VALKEY_CLOSE_CONNECTION = True
        await cache.aset("f", "1")
        await cache.aclose()

    async def test_close_client(self, cache: AsyncValkeyCache, mocker: MockerFixture):
        mock = mocker.patch.object(cache.client, "close")

        await cache.close()
        assert mock.called

    async def test_ttl(self, cache: AsyncValkeyCache):
        await cache.aset("foo", "bar", 10)
        ttl = await cache.attl("foo")

        if isinstance(cache.client, AsyncHerdClient):
            assert pytest.approx(ttl) == 12
        else:
            assert pytest.approx(ttl) == 10

        # Test ttl None
        await cache.aset("foo", "foo", timeout=None)
        ttl = await cache.attl("foo")
        assert ttl is None

        # Test ttl with expired key
        await cache.aset("foo", "foo", timeout=-1)
        ttl = await cache.attl("foo")
        assert ttl == 0

        # Test ttl with not existent key
        ttl = await cache.attl("not-existent-key")
        assert ttl == 0

    async def test_pttl(self, cache: AsyncValkeyCache):
        # Test pttl
        await cache.aset("foo", "bar", 10)
        ttl = await cache.apttl("foo")

        # delta is set to 10 as precision error causes tests to fail
        if isinstance(cache.client, AsyncHerdClient):
            assert pytest.approx(ttl, 10) == 12000
        else:
            assert pytest.approx(ttl, 10) == 10000

        # Test pttl with float value
        await cache.aset("foo", "bar", 5.5)
        ttl = await cache.apttl("foo")

        if isinstance(cache.client, AsyncHerdClient):
            assert pytest.approx(ttl, 10) == 7500
        else:
            assert pytest.approx(ttl, 10) == 5500

        # Test pttl None
        await cache.aset("foo", "foo", timeout=None)
        ttl = await cache.apttl("foo")
        assert ttl is None

        # Test pttl with expired key
        await cache.aset("foo", "foo", timeout=-1)
        ttl = await cache.apttl("foo")
        assert ttl == 0

        # Test pttl with not existent key
        ttl = await cache.apttl("not-existent-key")
        assert ttl == 0

    async def test_persist(self, cache: AsyncValkeyCache):
        await cache.aset("foo", "bar", timeout=20)
        assert await cache.apersist("foo") is True

        ttl = await cache.attl("foo")
        assert ttl is None
        assert await cache.apersist("not-existent-key") is False

    async def test_expire(self, cache: AsyncValkeyCache):
        await cache.aset("foo", "bar", timeout=None)
        assert await cache.aexpire("foo", 20) is True
        ttl = await cache.attl("foo")
        assert pytest.approx(ttl) == 20
        assert await cache.aexpire("not-existent-key", 20) is False

    async def test_expire_with_default_timeout(self, cache: AsyncValkeyCache):
        await cache.aset("foo", "bar", timeout=None)
        assert await cache.aexpire("foo", DEFAULT_TIMEOUT) is True
        assert await cache.aexpire("not-existent-key", DEFAULT_TIMEOUT) is False

    async def test_pexpire(self, cache: AsyncValkeyCache):
        await cache.aset("foo", "bar", timeout=None)
        assert await cache.apexpire("foo", 20500) is True
        ttl = await cache.apttl("foo")
        # delta is set to 10 as precision error causes tests to fail
        assert pytest.approx(ttl, 10) == 20500
        assert await cache.apexpire("not-existent-key", 20500) is False

    async def test_pexpire_with_default_timeout(self, cache: AsyncValkeyCache):
        await cache.aset("foo", "bar", timeout=None)
        assert await cache.apexpire("foo", DEFAULT_TIMEOUT) is True
        assert await cache.apexpire("not-existent-key", DEFAULT_TIMEOUT) is False

    async def test_pexpire_at(self, cache: AsyncValkeyCache):
        # Test settings expiration time 1 hour ahead by datetime.
        await cache.aset("foo", "bar", timeout=None)
        expiration_time = datetime.datetime.now() + timedelta(hours=1)
        assert await cache.apexpire_at("foo", expiration_time) is True
        ttl = await cache.apttl("foo")
        assert pytest.approx(ttl, 10) == timedelta(hours=1).total_seconds()

        # Test settings expiration time 1 hour ahead by Unix timestamp.
        await cache.aset("foo", "bar", timeout=None)
        expiration_time = datetime.datetime.now() + timedelta(hours=2)
        assert (
            await cache.apexpire_at("foo", int(expiration_time.timestamp() * 1000))
            is True
        )
        ttl = await cache.apttl("foo")
        assert pytest.approx(ttl, 10) == timedelta(hours=2).total_seconds() * 1000

        # Test settings expiration time 1 hour in the past, which effectively
        # deletes the key.
        expiration_time = datetime.datetime.now() - timedelta(hours=2)
        assert await cache.apexpire_at("foo", expiration_time) is True
        value = await cache.aget("foo")
        assert value is None

        expiration_time = datetime.datetime.now() + timedelta(hours=2)
        assert await cache.apexpire_at("not-existent-key", expiration_time) is False

    async def test_expire_at(self, cache: AsyncValkeyCache):
        # Test settings expiration time 1 hour ahead by datetime.
        await cache.aset("foo", "bar", timeout=None)
        expiration_time = datetime.datetime.now() + timedelta(hours=1)
        assert await cache.aexpire_at("foo", when=expiration_time) is True
        ttl = await cache.attl("foo")
        assert pytest.approx(ttl, 1) == timedelta(hours=1).total_seconds()

        # Test settings expiration time 1 hour ahead by Unix timestamp.
        await cache.aset("foo", "bar", timeout=None)
        expiration_time = datetime.datetime.now() + timedelta(hours=2)
        assert await cache.aexpire_at("foo", int(expiration_time.timestamp())) is True
        ttl = await cache.attl("foo")
        assert pytest.approx(ttl, 1) == timedelta(hours=1).total_seconds() * 2

        # Test settings expiration time 1 hour in the past, which effectively
        # deletes the key.
        expiration_time = datetime.datetime.now() - timedelta(hours=2)
        assert await cache.aexpire_at("foo", expiration_time) is True
        value = await cache.aget("foo")
        assert value is None

        expiration_time = datetime.datetime.now() + timedelta(hours=2)
        assert await cache.aexpire_at("not-existent-key", expiration_time) is False

    async def test_lock(self, cache: AsyncValkeyCache):
        lock = await cache.aget_lock("foobar")
        await lock.acquire(blocking=True)

        assert await cache.ahas_key("foobar")
        await lock.release()
        assert not await cache.ahas_key("foobar")

    @pytest.mark.filterwarnings("ignore")
    async def test_lock_released_by_thread(self, cache: AsyncValkeyCache):
        lock = await cache.lock("foobar", thread_local=False)
        await lock.acquire(blocking=True)

        async def release_lock(lock_):
            await lock_.release()

        t = threading.Thread(target=anyio.run, args=[release_lock, lock])
        t.start()
        t.join()

        assert not await cache.has_key("foobar")

    async def test_iter_keys(self, cache: AsyncValkeyCache):
        await cache.aset("foo1", 1)
        await cache.aset("foo2", 1)
        await cache.aset("foo3", 1)

        # Test simple result
        result = set()
        async with contextlib.aclosing(cache.aiter_keys("foo*")) as keys:
            async for k in keys:
                result.add(k)
        assert result == {"foo1", "foo2", "foo3"}

    async def test_iter_keys_itersize(self, cache: AsyncValkeyCache):
        await cache.aset("foo1", 1)
        await cache.aset("foo2", 1)
        await cache.aset("foo3", 1)

        # Test limited result
        result = []
        async with contextlib.aclosing(cache.aiter_keys("foo*", itersize=2)) as keys:
            async for k in keys:
                result.append(k)
        assert len(result) == 3

    async def test_iter_keys_generator(self, cache: AsyncValkeyCache):
        await cache.aset("foo1", 1)
        await cache.aset("foo2", 1)
        await cache.aset("foo3", 1)

        # Test generator object
        result = cache.aiter_keys("foo*")
        next_value = anext(result)  # noqa: F821
        assert await next_value is not None

    async def test_primary_replica_switching(self, cache: AsyncValkeyCache):
        cache = cast(AsyncValkeyCache, caches["sample"])
        client = cache.client
        client._server = ["foo", "bar"]
        client._clients = ["Foo", "Bar"]

        assert await client.get_client(write=True) == "Foo"
        assert await client.get_client(write=False) == "Bar"

    async def test_primary_replica_switching_with_index(self, cache: AsyncValkeyCache):
        cache = cast(AsyncValkeyCache, caches["sample"])
        client = cache.client
        client._server = ["foo", "bar"]
        client._clients = ["Foo", "Bar"]

        assert await client.get_client_with_index(write=True) == ("Foo", 0)
        assert await client.get_client_with_index(write=False) == ("Bar", 1)

    async def test_touch_zero_timeout(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", 222, timeout=10)

        assert await cache.atouch("test_key", 0) is True
        res = await cache.aget("test_key")
        assert res is None

    async def test_touch_positive_timeout(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", 222, timeout=10)

        assert await cache.atouch("test_key", 2) is True
        assert await cache.aget("test_key") == 222
        await anyio.sleep(3)
        assert await cache.aget("test_key") is None

    async def test_touch_negative_timeout(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", 222, timeout=10)

        assert await cache.atouch("test_key", -1) is True
        res = await cache.aget("test_key")
        assert res is None

    async def test_touch_missed_key(self, cache: AsyncValkeyCache):
        assert await cache.atouch("test_key_does_not_exist", 1) is False

    async def test_touch_forever(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", "foo", timeout=1)
        result = await cache.atouch("test_key", None)
        assert result is True
        assert await cache.attl("test_key") is None
        await anyio.sleep(2)
        assert await cache.aget("test_key") == "foo"

    async def test_touch_forever_nonexistent(self, cache: AsyncValkeyCache):
        result = await cache.atouch("test_key_does_not_exist", None)
        assert result is False

    async def test_touch_default_timeout(self, cache: AsyncValkeyCache):
        await cache.aset("test_key", "foo", timeout=1)
        result = await cache.atouch("test_key")
        assert result is True
        await anyio.sleep(2)
        assert await cache.aget("test_key") == "foo"

    async def test_clear(self, cache: AsyncValkeyCache):
        await cache.aset("foo", "bar")
        value_from_cache = await cache.aget("foo")
        assert value_from_cache == "bar"
        await cache.aclear()
        value_from_cache_after_clear = await cache.aget("foo")
        assert value_from_cache_after_clear is None

    async def test_hset(self, cache: AsyncValkeyCache):
        assert await cache.ahset("foo_hash1", "foo1", "bar1") == 1
        await cache.ahset("foo_hash1", "foo2", "bar2")
        assert await cache.ahlen("foo_hash1") == 2
        assert await cache.ahexists("foo_hash1", "foo1")
        assert await cache.ahexists("foo_hash1", "foo2")

    async def test_hdel(self, cache: AsyncValkeyCache):
        await cache.ahset("foo_hash2", "foo1", "bar1")
        await cache.ahset("foo_hash2", "foo2", "bar2")
        assert await cache.ahlen("foo_hash2") == 2
        deleted_count = await cache.ahdel("foo_hash2", "foo1")
        assert deleted_count == 1
        assert await cache.ahlen("foo_hash2") == 1
        assert not await cache.ahexists("foo_hash2", "foo1")
        assert await cache.ahexists("foo_hash2", "foo2")

    async def test_hlen(self, cache: AsyncValkeyCache):
        assert await cache.ahlen("foo_hash3") == 0
        await cache.ahset("foo_hash3", "foo1", "bar1")
        assert await cache.ahlen("foo_hash3") == 1
        await cache.ahset("foo_hash3", "foo2", "bar2")
        assert await cache.ahlen("foo_hash3") == 2

    async def test_hkeys(self, cache: AsyncValkeyCache):
        await cache.ahset("foo_hash4", "foo1", "bar1")
        await cache.ahset("foo_hash4", "foo2", "bar2")
        await cache.ahset("foo_hash4", "foo3", "bar3")
        keys = await cache.ahkeys("foo_hash4")
        assert len(keys) == 3
        for i in range(len(keys)):
            assert keys[i] == f"foo{i + 1}"

    async def test_hexists(self, cache: AsyncValkeyCache):
        await cache.ahset("foo_hash5", "foo1", "bar1")
        assert await cache.ahexists("foo_hash5", "foo1")
        assert not await cache.ahexists("foo_hash5", "foo")

    async def test_sadd(self, cache: AsyncValkeyCache):
        assert await cache.asadd("foo", "bar") == 1
        assert await cache.asmembers("foo") == {"bar"}

    async def test_sadd_int(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", 1)
        assert await cache.asmembers("foo") == {1}
        raw_client = await cache.client._get_client(write=False, client=None)
        assert await raw_client.smembers(":1:foo") == [b"1"]

    async def test_sadd_float(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", 1.2)
        assert await cache.asmembers("foo") == {1.2}
        raw_client = await cache.client._get_client(write=False, client=None)
        assert await raw_client.smembers(":1:foo") == [b"1.2"]

    async def test_scard(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar", "bar2")
        assert await cache.ascard("foo") == 2

    async def test_sdiff(self, cache: AsyncValkeyCache):
        await cache.asadd("foo1", "bar1", "bar2")
        await cache.asadd("foo2", "bar2", "bar3")
        assert await cache.asdiff("foo1", "foo2") == {"bar1"}

    async def test_sdiffstore(self, cache: AsyncValkeyCache):
        await cache.asadd("foo1", "bar1", "bar2")
        await cache.asadd("foo2", "bar2", "bar3")
        assert await cache.asdiffstore("foo3", "foo1", "foo2") == 1
        assert await cache.asmembers("foo3") == {"bar1"}

    async def test_sdiffstore_with_keys_version(self, cache: AsyncValkeyCache):
        await cache.asadd("foo1", "bar1", "bar2", version=2)
        await cache.asadd("foo2", "bar2", "bar3", version=2)
        assert await cache.asdiffstore("foo3", "foo1", "foo2", version_keys=2) == 1
        assert await cache.asmembers("foo3") == {"bar1"}

    async def test_sdiffstore_with_different_keys_versions_without_initial_set_in_version(
        self, cache: AsyncValkeyCache
    ):
        await cache.asadd("foo1", "bar1", "bar2", version=1)
        await cache.asadd("foo2", "bar2", "bar3", version=2)
        assert await cache.asdiffstore("foo3", "foo1", "foo2", version_keys=2) == 0

    async def test_sdiffstore_with_different_keys_versions_with_initial_set_in_version(
        self, cache: AsyncValkeyCache
    ):
        await cache.asadd("foo1", "bar1", "bar2", version=2)
        await cache.asadd("foo2", "bar2", "bar3", version=1)
        assert await cache.asdiffstore("foo3", "foo1", "foo2", version_keys=2) == 2

    async def test_sinter(self, cache: AsyncValkeyCache):
        await cache.asadd("foo1", "bar1", "bar2")
        await cache.asadd("foo2", "bar2", "bar3")
        assert await cache.asinter("foo1", "foo2") == {"bar2"}

    async def test_interstore(self, cache: AsyncValkeyCache):
        await cache.asadd("foo1", "bar1", "bar2")
        await cache.asadd("foo2", "bar2", "bar3")
        assert await cache.asinterstore("foo3", "foo1", "foo2") == 1
        assert await cache.asmembers("foo3") == {"bar2"}

    async def test_sismember_str(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar")
        assert await cache.asismember("foo", "bar") is True
        assert await cache.asismember("foo", "bar2") is False

    async def test_sismember_int(self, cache: AsyncValkeyCache):
        await cache.asadd("baz", 3)
        assert await cache.asismember("baz", 3) is True
        assert await cache.asismember("baz", 2) is False

    async def test_smove(self, cache: AsyncValkeyCache):
        await cache.asadd("foo1", "bar1", "bar2")
        await cache.asadd("foo2", "bar2", "bar3")
        assert await cache.asmove("foo1", "foo2", "bar1") is True
        assert await cache.asmove("foo1", "foo2", "bar4") is False
        assert await cache.asmembers("foo1") == {"bar2"}
        assert await cache.asmembers("foo2") == {"bar1", "bar2", "bar3"}

    async def test_spop_default_count(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar1", "bar2")
        assert await cache.aspop("foo") in {"bar1", "bar2"}
        assert await cache.asmembers("foo") in [{"bar1"}, {"bar2"}]

    async def test_spop(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar1", "bar2")
        assert await cache.aspop("foo", 1) in [{"bar1"}, {"bar2"}]
        assert await cache.asmembers("foo") in [{"bar1"}, {"bar2"}]

    async def test_srandmember_default_count(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar1", "bar2")
        assert await cache.asrandmember("foo") in {"bar1", "bar2"}

    async def test_srandmember(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar1", "bar2")
        assert await cache.asrandmember("foo", 1) in [["bar1"], ["bar2"]]

    async def test_srem(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar1", "bar2")
        assert await cache.asrem("foo", "bar1") == 1
        assert await cache.asrem("foo", "bar3") == 0

    async def test_sscan(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar1", "bar2")
        items = await cache.asscan("foo")
        assert items == {"bar1", "bar2"}

    async def test_sscan_with_match(self, cache: AsyncValkeyCache):
        if cache.client._has_compression_enabled():
            pytest.skip("Compression is enabled, sscan with match is not supported")
        await cache.asadd("foo", "bar1", "bar2", "zoo")
        items = await cache.asscan("foo", match="zoo")
        assert items == {"zoo"}

    async def test_sscan_iter(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar1", "bar2")
        items = set()
        async with contextlib.aclosing(cache.asscan_iter("foo")) as values:
            async for value in values:
                items.add(value)
        assert items == {"bar1", "bar2"}

    async def test_sscan_iter_with_match(self, cache: AsyncValkeyCache):
        if cache.client._has_compression_enabled():
            pytest.skip(
                "Compression is enabled, sscan_iter with match is not supported"
            )
        await cache.asadd("foo", "bar1", "bar2", "zoo")

        items = set()
        async with contextlib.aclosing(
            cache.asscan_iter("foo", match="bar*")
        ) as values:
            async for value in values:
                items.add(value)

        assert set(items) == {"bar1", "bar2"}

    async def test_smismember(self, cache: AsyncValkeyCache):
        await cache.asadd("foo", "bar1", "bar2", "bar3")
        assert await cache.asmismember("foo", "bar1", "bar2", "xyz") == [
            True,
            True,
            False,
        ]

    async def test_sunion(self, cache: AsyncValkeyCache):
        await cache.asadd("foo1", "bar1", "bar2")
        await cache.asadd("foo2", "bar2", "bar3")
        assert await cache.asunion("foo1", "foo2") == {"bar1", "bar2", "bar3"}

    async def test_sunionstore(self, cache: AsyncValkeyCache):
        await cache.asadd("foo1", "bar1", "bar2")
        await cache.asadd("foo2", "bar2", "bar3")
        assert await cache.asunionstore("foo3", "foo1", "foo2") == 3
        assert await cache.asmembers("foo3") == {"bar1", "bar2", "bar3"}

    async def test_make_key(self, cache: AsyncValkeyCache):
        assert cache.make_key("key", version=1, prefix="prefix") == "prefix:1:key"

    async def test_make_pattern(self, cache: AsyncValkeyCache):
        assert (
            cache.make_pattern("key_*", version=1, prefix="prefix") == "prefix:1:key_*"
        )



import pytest

from django_valkey import pool

pytestmark = pytest.mark.anyio


@pytest.mark.parametrize(
    "connection_string",
    [
        "unix://tmp/foo.bar?db=1",
        "valkey://localhost/2",
        "valkeys://localhost:3333?db=2",
    ],
)
async def test_connection_strings(connection_string: str):
    cf = pool.get_connection_factory(
        options={
            "CONNECTION_FACTORY": "django_valkey.async_cache.pool.AsyncConnectionFactory"
        }
    )
    res = cf.make_connection_params(connection_string)
    assert res["url"] == connection_string



from valkey import Valkey

from django_valkey.base import (
    BaseValkeyCache,
    BackendCommands,
    decorate_all_methods,
    omit_exception,
)
from django_valkey.client import DefaultClient


@decorate_all_methods(omit_exception)
class DecoratedBackendCommands(BackendCommands):
    pass


class ValkeyCache(BaseValkeyCache[DefaultClient, Valkey], DecoratedBackendCommands):
    DEFAULT_CLIENT_CLASS = "django_valkey.client.DefaultClient"



SECRET_KEY = "django_tests_secret_key"
CACHES = {
    "default": {
        "BACKEND": "django_valkey.cache.ValkeyCache",
        "LOCATION": ["valkey://127.0.0.1:6379?db=1", "valkey://127.0.0.1:6379?db=1"],
        "OPTIONS": {"CLIENT_CLASS": "django_valkey.client.DefaultClient"},
    },
    "doesnotexist": {
        "BACKEND": "django_valkey.cache.ValkeyCache",
        "LOCATION": "valkey://127.0.0.1:56379?db=1",
        "OPTIONS": {"CLIENT_CLASS": "django_valkey.client.DefaultClient"},
    },
    "sample": {
        "BACKEND": "django_valkey.cache.ValkeyCache",
        "LOCATION": "valkey://127.0.0.1:6379:1,valkey://127.0.0.1:6379:1",
        "OPTIONS": {"CLIENT_CLASS": "django_valkey.client.DefaultClient"},
    },
    "with_prefix": {
        "BACKEND": "django_valkey.cache.ValkeyCache",
        "LOCATION": "valkey://127.0.0.1:6379?db=1",
        "OPTIONS": {"CLIENT_CLASS": "django_valkey.client.DefaultClient"},
        "KEY_PREFIX": "test-prefix",
    },
}

# Include `django.contrib.auth` and `django.contrib.contenttypes` for mypy /
# django-stubs.

# See:
# - https://github.com/typeddjango/django-stubs/issues/318
# - https://github.com/typeddjango/django-stubs/issues/534
INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
]

USE_TZ = False

ROOT_URLCONF = "tests.settings.urls"



from valkey.asyncio.client import Valkey as AValkey

from django_valkey.base import (
    BaseValkeyCache,
    AsyncBackendCommands,
    decorate_all_methods,
    omit_exception_async,
)
from django_valkey.async_cache.client.default import AsyncDefaultClient


@decorate_all_methods(omit_exception_async)
class DecoratedAsyncBackendCommands(AsyncBackendCommands):
    pass


class AsyncValkeyCache(
    BaseValkeyCache[AsyncDefaultClient, AValkey], DecoratedAsyncBackendCommands
):
    DEFAULT_CLIENT_CLASS = "django_valkey.async_cache.client.default.AsyncDefaultClient"
    is_async = True



import copy
from typing import cast

import pytest
from pytest import LogCaptureFixture
from pytest_django.fixtures import SettingsWrapper

from django.core.cache import caches, cache as default_cache

from valkey.exceptions import ConnectionError

from django_valkey.cache import ValkeyCache
from django_valkey.client import ShardClient, HerdClient, DefaultClient
from django_valkey.cluster_cache.client import DefaultClusterClient


def make_key(key: str, prefix: str, version: str) -> str:
    return f"{prefix}#{version}#{key}"


def reverse_key(key: str) -> str:
    return key.split("#", 2)[2]


methods_with_no_parameters = {"clear", "close"}

methods_with_one_required_parameters = {
    "decr",
    "delete",
    "delete_many",
    "delete_pattern",
    "get",
    "get_many",
    "has_key",
    "hkeys",
    "hlen",
    "incr",
    "incr_version",
    "decr_version",
    "keys",
    "lock",
    "mget",
    "pttl",
    "persist",
    "scard",
    "sdiff",
    "sinter",
    "smembers",
    "spop",
    "srandmember",
    "sscan",
    "sunion",
    "touch",
    "ttl",
}
methods_with_two_required_parameters = {
    "add",
    "expire",
    "expire_at",
    "hdel",
    "hexists",
    "pexpire",
    "pexpire_at",
    "sadd",
    "sdiffstore",
    "set",
    "sinterstore",
    "sismember",
    "smismember",
    "srem",
    "sunionstore",
}
methods_with_three_required_parameters = {"hset", "smove"}
methods_taking_dictionary = {"mset", "set_many"}
iter_methods = {
    "iter_keys",
    "sscan_iter",
}

no_shard_methods = {
    "mget",
    "mset",
    "iter_keys",
}
no_herd_method = {
    "incr",
    "decr",
}


@pytest.mark.skipif(
    isinstance(default_cache.client, DefaultClusterClient),
    reason="cluster client doesn't support ignore exception",
)
class TestDjangoValkeyOmitException:
    @pytest.fixture
    def conf_cache(self, settings: SettingsWrapper):
        caches_setting = copy.deepcopy(settings.CACHES)
        settings.CACHES = caches_setting
        return caches_setting

    @pytest.fixture
    def conf_cache_to_ignore_exception(self, settings: SettingsWrapper, conf_cache):
        conf_cache["doesnotexist"]["OPTIONS"]["IGNORE_EXCEPTIONS"] = True
        conf_cache["doesnotexist"]["OPTIONS"]["LOG_IGNORED_EXCEPTIONS"] = True
        settings.DJANGO_VALKEY_IGNORE_EXCEPTIONS = True
        settings.DJANGO_VALKEY_LOG_IGNORED_EXCEPTIONS = True

    @pytest.fixture
    def ignore_exceptions_cache(self, conf_cache_to_ignore_exception) -> ValkeyCache:
        return cast(ValkeyCache, caches["doesnotexist"])

    def test_methods_with_no_argument_omit_exception(
        self, ignore_exceptions_cache: ValkeyCache, subtests
    ):
        for m in methods_with_no_parameters:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                method()

    def test_methods_with_one_argument_omit_exception(
        self, ignore_exceptions_cache: ValkeyCache, subtests
    ):
        for m in methods_with_one_required_parameters:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                if (
                    isinstance(default_cache.client, ShardClient)
                    and m in no_shard_methods
                ):
                    pytest.skip(f"shard client doesn't support {m}")
                elif (
                    isinstance(default_cache.client, HerdClient) and m in no_herd_method
                ):
                    pytest.skip(f"herd client doesn't support {m}")
                method("abc")

    def test_methods_with_two_argument_omit_exception(
        self, ignore_exceptions_cache: ValkeyCache, subtests
    ):
        for m in methods_with_two_required_parameters:
            method = getattr(ignore_exceptions_cache, m)
            if isinstance(default_cache.client, ShardClient) and m in no_shard_methods:
                pytest.skip(f"shard client doesn't support {m}")
            with subtests.test(method=method):
                method("abc", 1)

    @pytest.mark.skipif(
        not isinstance(default_cache.client, DefaultClient),
        reason="not supported by none default clients",
    )
    def test_methods_with_three_argument_omit_exception(
        self, ignore_exceptions_cache: ValkeyCache, subtests
    ):
        for m in methods_with_three_required_parameters:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                method("abc", "def", "ghi")

    def test_methods_taking_dictionary(
        self, ignore_exceptions_cache: ValkeyCache, subtests
    ):
        for m in methods_taking_dictionary:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                if (
                    isinstance(default_cache.client, ShardClient)
                    and m in no_shard_methods
                ):
                    pytest.skip(f"shard client doesn't support {m}")
                method({"abc": "def"})

    def test_iterator_methods(self, ignore_exceptions_cache: ValkeyCache, subtests):
        for m in iter_methods:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                if (
                    isinstance(default_cache.client, ShardClient)
                    and m in no_shard_methods
                ):
                    pytest.skip(f"shard client doesn't support {m}")
                for _ in method("abc"):
                    pass

    def test_get_django_omit_exceptions_many_returns_default_arg(
        self,
        ignore_exceptions_cache: ValkeyCache,
    ):
        assert ignore_exceptions_cache._ignore_exceptions is True
        assert ignore_exceptions_cache.get_many(["key1", "key2", "key3"]) == {}

    def test_get_django_omit_exceptions(
        self, caplog: LogCaptureFixture, ignore_exceptions_cache: ValkeyCache
    ):
        assert ignore_exceptions_cache._ignore_exceptions is True
        assert ignore_exceptions_cache._log_ignored_exceptions is True

        assert ignore_exceptions_cache.get("key") is None
        assert ignore_exceptions_cache.get("key", "default") == "default"
        assert ignore_exceptions_cache.get("key", default="default") == "default"

        assert len(caplog.records) == 3
        assert all(
            record.levelname == "ERROR" and record.msg == "Exception ignored"
            for record in caplog.records
        )

    def test_get_django_omit_exceptions_priority_1(self, settings: SettingsWrapper):
        caches_setting = copy.deepcopy(settings.CACHES)
        caches_setting["doesnotexist"]["OPTIONS"]["IGNORE_EXCEPTIONS"] = True
        settings.CACHES = caches_setting
        settings.DJANGO_VALKEY_IGNORE_EXCEPTIONS = False
        cache = cast(ValkeyCache, caches["doesnotexist"])
        assert cache._ignore_exceptions is True
        assert cache.get("key") is None

    def test_get_django_omit_exceptions_priority_2(self, settings: SettingsWrapper):
        caches_setting = copy.deepcopy(settings.CACHES)
        caches_setting["doesnotexist"]["OPTIONS"]["IGNORE_EXCEPTIONS"] = False
        settings.CACHES = caches_setting
        settings.DJANGO_VALKEY_IGNORE_EXCEPTIONS = True
        cache = cast(ValkeyCache, caches["doesnotexist"])
        assert cache._ignore_exceptions is False
        with pytest.raises(ConnectionError):
            cache.get("key")

    def test_error_raised_when_ignore_is_not_set(self, conf_cache):
        cache = caches["doesnotexist"]
        assert cache._ignore_exceptions is False
        assert cache._log_ignored_exceptions is False
        with pytest.raises(ConnectionError):
            cache.get("key")


class TestDjangoValkeyCacheEscapePrefix:
    def test_delete_pattern(
        self, key_prefix_cache: ValkeyCache, with_prefix_cache: ValkeyCache
    ):
        key_prefix_cache.set("a", "1")
        with_prefix_cache.set("b", "2")
        key_prefix_cache.delete_pattern("*")
        assert key_prefix_cache.has_key("a") is False
        assert with_prefix_cache.get("b") == "2"

    def test_iter_keys(
        self, key_prefix_cache: ValkeyCache, with_prefix_cache: ValkeyCache
    ):
        if isinstance(key_prefix_cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support iter_keys")

        key_prefix_cache.set("a", "1")
        with_prefix_cache.set("b", "2")
        assert list(key_prefix_cache.iter_keys("*")) == ["a"]

    @pytest.mark.skipif(
        isinstance(default_cache.client, DefaultClusterClient),
        reason="cluster client doesn't support ignore exception",
    )
    def test_keys(self, key_prefix_cache: ValkeyCache, with_prefix_cache: ValkeyCache):
        key_prefix_cache.set("a", "1")
        with_prefix_cache.set("b", "2")
        keys = key_prefix_cache.keys("*")
        assert "a" in keys
        assert "b" not in keys


@pytest.mark.skipif(
    isinstance(default_cache.client, DefaultClusterClient),
    reason="cluster client doesn't support ignore exception",
)
def test_custom_key_function(cache: ValkeyCache, settings: SettingsWrapper):
    caches_setting = copy.deepcopy(settings.CACHES)
    caches_setting["default"]["KEY_FUNCTION"] = "tests.test_cache_options.make_key"
    caches_setting["default"]["REVERSE_KEY_FUNCTION"] = (
        "tests.test_cache_options.reverse_key"
    )
    settings.CACHES = caches_setting

    if isinstance(cache.client, ShardClient):
        pytest.skip("ShardClient doesn't support get_client")

    for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
        cache.set(key, "foo")

    res = cache.delete_pattern("*foo-a*")
    assert bool(res) is True

    keys = cache.keys("foo*")
    assert set(keys) == {"foo-bb", "foo-bc"}
    # ensure our custom function was actually called
    assert {k.decode() for k in cache.client.get_client(write=False).keys("*")} == (
        {"#1#foo-bc", "#1#foo-bb"}
    )



from valkey.cluster import ValkeyCluster

from django_valkey.base import BaseValkeyCache, BackendCommands
from django_valkey.cluster_cache.client import DefaultClusterClient


class ClusterCommands:
    def msetnx(self: "ClusterValkeyCache", *args, **kwargs):
        return self.client.msetnx(*args, **kwargs)

    def mget_nonatomic(self: "ClusterValkeyCache", *args, **kwargs):
        return self.client.mget_nonatomic(*args, **kwargs)

    def mset_nonatomic(self: "ClusterValkeyCache", *args, **kwargs):
        return self.client.mset_nonatomic(*args, **kwargs)

    def readonly(self: "ClusterValkeyCache", *args, **kwargs):
        return self.client.readonly(*args, **kwargs)

    def readwrite(self: "ClusterValkeyCache", *args, **kwargs):
        return self.client.readwrite(*args, **kwargs)

    def keyslot(self: "ClusterValkeyCache", *args, **kwargs):
        return self.client.keyslot(*args, **kwargs)

    def flushall(self: "ClusterValkeyCache", *args, **kwargs):
        return self.client.flushall(*args, **kwargs)

    def invalidate_key_from_cache(self: "ClusterValkeyCache", *args, **kwargs):
        return self.client.invalidate_key_from_cache(*args, **kwargs)


class ClusterValkeyCache(
    BaseValkeyCache[DefaultClusterClient, ValkeyCluster],
    ClusterCommands,
    BackendCommands,
):
    DEFAULT_CLIENT_CLASS = "django_valkey.cluster_cache.client.DefaultClusterClient"



from collections.abc import Iterable

import pytest
from pytest_django.fixtures import SettingsWrapper
from pytest_mock import MockerFixture

from django.core.cache import DEFAULT_CACHE_ALIAS, cache as default_cache

from django_valkey.cache import ValkeyCache
from django_valkey.client import DefaultClient, ShardClient


@pytest.fixture
def cache_client(cache: ValkeyCache) -> Iterable[DefaultClient]:
    client = cache.client
    client.set("TestClientClose", 0)
    yield client
    client.delete("TestClientClose")


class TestClientClose:
    def test_close_client_disconnect_default(
        self, cache_client: DefaultClient, mocker: MockerFixture
    ):
        mock = mocker.patch.object(cache_client.connection_factory, "disconnect")

        cache_client.close()
        assert not mock.called

    def test_close_disconnect_settings(
        self,
        cache_client: DefaultClient,
        settings: SettingsWrapper,
        mocker: MockerFixture,
    ):
        mock = mocker.patch.object(cache_client.connection_factory, "disconnect")

        settings.DJANGO_VALKEY_CLOSE_CONNECTION = True
        cache_client.close()
        assert mock.called

    def test_close_disconnect_settings_cache(
        self,
        cache_client: DefaultClient,
        mocker: MockerFixture,
        settings: SettingsWrapper,
    ):
        mock = mocker.patch.object(cache_client.connection_factory, "disconnect")

        settings.CACHES[DEFAULT_CACHE_ALIAS]["OPTIONS"]["CLOSE_CONNECTION"] = True
        cache_client.set("TestClientClose", 0)

        cache_client.close()
        assert mock.called

    def test_close_disconnect_client_options(
        self, cache_client: DefaultClient, mocker: MockerFixture
    ):
        mock = mocker.patch.object(cache_client.connection_factory, "disconnect")

        cache_client._options["CLOSE_CONNECTION"] = True

        cache_client.close()
        assert mock.called


@pytest.mark.skipif(
    not isinstance(default_cache.client, DefaultClient), reason="shard only test"
)
class TestDefaultClient:
    def test_delete_pattern_calls_get_client_given_no_client(self, mocker):
        get_client_mock = mocker.patch(
            "django_valkey.base_client.ClientCommands.get_client"
        )
        mocker.patch("django_valkey.base_client.BaseClient.__init__", return_value=None)
        client = DefaultClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""

        client.delete_pattern(pattern="foo*")
        get_client_mock.assert_called_once_with(write=True, tried=None)

    def test_delete_pattern_calls_make_pattern(self, mocker):
        make_pattern_mock = mocker.patch(
            "django_valkey.base_client.BaseClient.make_pattern"
        )
        get_client_mock = mocker.patch(
            "django_valkey.base_client.ClientCommands.get_client",
            return_value=mocker.Mock(),
        )
        mocker.patch("django_valkey.base_client.BaseClient.__init__", return_value=None)
        client = DefaultClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""
        get_client_mock.return_value.scan_iter.return_value = []

        client.delete_pattern(pattern="foo*")

        kwargs = {"version": None, "prefix": None}
        make_pattern_mock.assert_called_once_with("foo*", **kwargs)

    def test_delete_pattern_calls_scan_iter_with_count_if_itersize_given(self, mocker):
        mocker.patch("django_valkey.base_client.BaseClient.__init__", return_value=None)
        get_client_mock = mocker.patch(
            "django_valkey.base_client.ClientCommands.get_client",
            return_value=mocker.Mock(),
        )
        make_pattern_mock = mocker.patch(
            "django_valkey.base_client.BaseClient.make_pattern"
        )
        client = DefaultClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""
        get_client_mock.return_value.scan_iter.return_value = []

        client.delete_pattern(pattern="foo*", itersize=90210)

        get_client_mock.return_value.scan_iter.assert_called_once_with(
            count=90210, match=make_pattern_mock.return_value
        )

    def test_delete_pattern_calls_pipeline_delete_and_execute(self, mocker):
        mocker.patch("django_valkey.base_client.BaseClient.__init__", return_value=None)
        get_client_mock = mocker.patch(
            "django_valkey.base_client.ClientCommands.get_client",
            return_value=mocker.Mock(),
        )
        mocker.patch("django_valkey.base_client.BaseClient.make_pattern")
        client = DefaultClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""
        get_client_mock.return_value.scan_iter.return_value = [":1:foo", ":1:foo-a"]
        get_client_mock.return_value.pipeline.return_value = mocker.Mock()
        get_client_mock.return_value.pipeline.return_value.delete = mocker.Mock()
        get_client_mock.return_value.pipeline.return_value.execute = mocker.Mock()

        client.delete_pattern(pattern="foo*")

        assert get_client_mock.return_value.pipeline.return_value.delete.call_count == 2
        get_client_mock.return_value.pipeline.return_value.delete.assert_has_calls(
            [mocker.call(":1:foo"), mocker.call(":1:foo-a")]
        )
        get_client_mock.return_value.pipeline.return_value.execute.assert_called_once()


@pytest.mark.skipif(
    not isinstance(default_cache.client, ShardClient), reason="shard only test"
)
class TestShardClient:
    CLIENT_METHODS_FOR_MOCK = {
        "add",
        "close",
        "expire",
        "get",
        "ttl",
        "touch",
        "sadd",
        "scan_iter",
        "spop",
    }

    @pytest.fixture
    def shard_cache(self):
        from django.core.cache import caches, ConnectionProxy

        cache = ConnectionProxy(caches, "default")
        yield cache
        cache.clear()

    @pytest.fixture
    def connection(self, mocker):
        connection = mocker.Mock()

        for m in self.CLIENT_METHODS_FOR_MOCK:
            setattr(connection, m, mocker.Mock(spec_set=()))

        connection.scan_iter.return_value = []
        connection.get.return_value = "this"
        connection.add.return_value = True
        connection.expire.return_value = False
        connection.close.return_value = None
        connection.ttl.return_value = 0
        connection.touch.return_value = True
        connection.sadd.return_value = 1
        connection.spop.return_value = {"this"}

        yield connection

    def test_delete_pattern_calls_scan_iter_with_count_if_itersize_given(
        self, connection, mocker
    ):
        mocker.patch(
            "django_valkey.client.sharded.ShardClient.__init__", return_value=None
        )
        make_pattern_mock = mocker.patch(
            "django_valkey.base_client.BaseClient.make_pattern"
        )
        client = ShardClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""

        client._server_dict = {"test": connection}

        client.delete_pattern(pattern="foo*", itersize=10)

        connection.scan_iter.assert_called_once_with(
            count=10, match=make_pattern_mock.return_value
        )

    def test_delete_pattern_calls_scan_iter(self, connection, mocker):
        mocker.patch(
            "django_valkey.client.sharded.ShardClient.__init__", return_value=None
        )
        make_pattern_mock = mocker.patch(
            "django_valkey.base_client.BaseClient.make_pattern"
        )
        client = ShardClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""
        client._server_dict = {"test": connection}

        client.delete_pattern(pattern="foo*")

        connection.scan_iter.assert_called_once_with(
            match=make_pattern_mock.return_value
        )

    def test_delete_pattern_calls_delete_for_given_keys(
        self, connection, cache, mocker
    ):
        mocker.patch(
            "django_valkey.client.sharded.ShardClient.__init__", return_value=None
        )
        mocker.patch("django_valkey.base_client.BaseClient.make_pattern")
        client = ShardClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""
        connection.scan_iter.return_value = [mocker.Mock(), mocker.Mock()]
        connection.delete.return_value = 0
        client._server_dict = {"test": connection}

        client.delete_pattern(pattern="foo*")

        connection.delete.assert_called_once_with(*connection.scan_iter.return_value)

    def test_shard_client_get_server_name_is_called(
        self, mocker, connection, shard_cache
    ):
        if not isinstance(shard_cache.client, ShardClient):
            pytest.skip("shard only test")
        client = ShardClient(
            server=shard_cache._server, params=shard_cache._params, backend=shard_cache
        )
        spy = mocker.spy(client, "get_server_name")

        client.add("that", "this")
        client.get("that")
        client.ttl("that")
        client.expire("that", 1)
        client.touch("that")
        client.sadd("key", "value1", "value2")
        client.spop("key")
        assert spy.call_count == 7



import sys
from django.conf import settings

from django_valkey.compressors.base import BaseCompressor

if sys.version_info >= (3, 14):
    from compression import zstd
else:
    from backports import zstd


class ZStdCompressor(BaseCompressor):
    """
    ZStdCompressor
    set with:
    CACHES = {
        "default": {
            # ...
            "OPTIONS": {
                "COMPRESSOR": "django_valkey.compressors.zstd.ZStdCompressor",
                }
            }
        }

    compression parameters:
    to set `level` use `CACHE_COMPRESS_LEVEL`
    to set `options` `COMPRESS_ZSTD_OPTIONS` in your settings.
    if `COMPRESSION_ZSTD_OPTIONS` is set, level won't be used

    to set `minimum size` set `CACHE_COMPRESS_MIN_LENGTH` in your settings, defaults to 15.

    to set `zstd_dict` use `COMPRESS_ZSTD_DICT` in your settings, it defaults to None.

    decompression parameters:
    to set `options` use `DECOMPRESS_ZSTD_OPTIONS` in your settings,
    if no value provided, it uses `COMPRESS_ZSTD_OPTIONS`.

    to set `zstd_dict` use `DECOMPRESS_ZSTD_DICT` in your settings,
    if no value is provided `COMPRESS_ZSTD_DICT` is used.

    """

    options: dict | None = getattr(settings, "COMPRESS_ZSTD_OPTIONS", None)
    decomp_options = getattr(settings, "DECOMPRESS_ZSTD_OPTIONS", None)
    zstd_dict = getattr(settings, "COMPRESS_ZSTD_DICT", None)
    decomp_zstd_dict = getattr(settings, "DECOMPRESS_ZSTD_DICT", None)

    def _compress(self, value: bytes) -> bytes:
        return zstd.compress(
            value,
            options=self.options,
            level=self.level or 1,
            zstd_dict=self.zstd_dict,
        )

    def _decompress(self, value: bytes) -> bytes:
        return zstd.decompress(
            value,
            zstd_dict=self.decomp_zstd_dict or self.zstd_dict,
            options=self.decomp_options or self.options,
        )



import os
import shutil

from invoke import task


@task
def devenv(c):
    clean(c)
    cmd = "docker compose --profile all up -d"
    c.run(cmd)


@task
def clean(c):
    if os.path.isdir("build"):
        shutil.rmtree("build")
    if os.path.isdir("dist"):
        shutil.rmtree("dist")

    c.run("docker compose --profile all rm -s -f")



import pytest

from django_valkey import pool


@pytest.mark.parametrize(
    "connection_string",
    [
        "unix://tmp/foo.bar?db=1",
        "valkey://localhost/2",
        "valkeys://localhost:3333?db=2",
    ],
)
def test_connection_strings(connection_string: str):
    cf = pool.get_connection_factory(
        options={"CONNECTION_FACTORY": "django_valkey.pool.ConnectionFactory"}
    )
    res = cf.make_connection_params(connection_string)
    assert res["url"] == connection_string



from django.core.cache import cache
from django.http import HttpResponse
from django.urls import path


async def async_view(request):
    res = await cache.get("a")
    return HttpResponse(res)


async def sync_view(request):
    res = cache.get("a")
    return HttpResponse(res)


urlpatterns = [
    path("async/", async_view),
    path("sync/", sync_view),
]



import pytest

from django.core import signals
from django.core.cache import close_caches

from django_valkey.base import close_async_caches
from django_valkey.async_cache.cache import AsyncValkeyCache


pytestmark = pytest.mark.anyio


class TestWithOldSignal:
    @pytest.fixture(autouse=True)
    def setup(self):
        signals.request_finished.disconnect(close_async_caches)
        signals.request_finished.connect(close_caches)
        yield
        signals.request_finished.disconnect(close_caches)
        signals.request_finished.connect(close_async_caches)

    def test_old_receiver_is_registered_and_new_receiver_unregistered(self, setup):
        sync_receivers, async_receivers = signals.request_finished._live_receivers(None)
        assert close_caches in sync_receivers
        assert close_async_caches not in async_receivers

    async def test_warning_output_when_request_finished(self, async_client):
        with pytest.warns(
            RuntimeWarning,
            match="coroutine 'AsyncBackendCommands.close' was never awaited",
        ) as record:
            await async_client.get("/async/")

        assert (
            str(record[0].message)
            == "coroutine 'AsyncBackendCommands.close' was never awaited"
        )

    async def test_manually_await_signal(self, recwarn):
        await signals.request_finished.asend(self.__class__)
        assert len(recwarn) == 1

        assert (
            str(recwarn[0].message)
            == "coroutine 'AsyncBackendCommands.close' was never awaited"
        )

    # TODO: find why garbage collector doesn't collect the coroutine when the method is
    # sync (even when gc is called manually, it doesn't collect)
    async def test_manually_call_signal(self):
        with pytest.warns(
            RuntimeWarning,
            match="coroutine 'AsyncBackendCommands.close' was never awaited",
        ) as record:
            signals.request_finished.send(self.__class__)
        assert len(record) == 1

        assert (
            str(record[0].message)
            == "coroutine 'AsyncBackendCommands.close' was never awaited"
        )


class TestWithNewSignal:
    async def test_warning_output_when_request_finished(self, async_client, recwarn):
        await async_client.get("/async/")

        assert len(recwarn) == 0

    async def test_manually_await_signal(self, recwarn):
        await signals.request_finished.asend(self.__class__)
        assert len(recwarn) == 0

    def test_manually_call_signal(self, recwarn):
        signals.request_finished.send(self.__class__)
        assert len(recwarn) == 0

    def test_receiver_is_registered_and_old_receiver_unregistered(self):
        sync_receivers, async_receivers = signals.request_finished._live_receivers(None)
        assert close_async_caches in async_receivers
        assert close_caches not in sync_receivers

    async def test_close_is_called_by_signal(self, mocker):
        close_spy = mocker.spy(AsyncValkeyCache, "close")
        await signals.request_finished.asend(self.__class__)
        assert close_spy.await_count == 1
        assert close_spy.call_count == 1



from valkey.asyncio import Valkey as AValkey

from django_valkey.base_client import BaseClient, AsyncClientCommands


class AsyncDefaultClient(BaseClient[AValkey], AsyncClientCommands[AValkey]):
    CONNECTION_FACTORY_PATH = "django_valkey.async_cache.pool.AsyncConnectionFactory"



import re
from collections import OrderedDict
from typing import Any, List, Dict

from valkey import Valkey
from valkey.typing import EncodableT, KeyT

from django_valkey.base_client import DEFAULT_TIMEOUT, _main_exceptions
from django_valkey.client.default import DefaultClient
from django_valkey.exceptions import ConnectionInterrupted
from django_valkey.hash_ring import HashRing

"""
supported methods:

`add()`
`get()`
`get_many()`
`set()`
`set_many()`
`has_key()`
`delete()`
`delete_many()`
`delete_pattern()`
`ttl()`
`pttl()`
`persist()`
`expire()`
`expire_at()`
`pexpire()`
`pexpire_at()`
`get_lock()`
`incr_version()`
`incr()`
`decr()`
`keys()`
`close()`
`touch()`
`clear()`
`sadd()`
`scard()`
`smembers()`
`smove()`
`srem()`
`sscan()`
`sscan_iter()`
`srandmember()`
`sismember()`
`spop()`
`smismember()`
"""


class ShardClient(DefaultClient):
    _findhash = re.compile(r".*\{(.*)\}.*", re.I)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not isinstance(self._server, (list, tuple, set)):
            self._server = [self._server]

        self._ring = HashRing(self._server)
        self._server_dict = self.connect()

    def connect(self, index: int = 0) -> dict:
        connection_dict = {}
        for name in self._server:
            connection_dict[name] = self.connection_factory.connect(name)
        return connection_dict

    def get_server_name(self, _key: KeyT) -> str:
        key = str(_key)
        g = self._findhash.match(key)
        if g is not None and len(g.groups()) > 0:
            key = g.groups()[0]
        return self._ring.get_node(key)

    def get_client(self, key: KeyT = None, **kwargs):
        name = self.get_server_name(key)
        return self._server_dict[name]

    def add(
        self,
        key: KeyT,
        value: EncodableT,
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_client(key=key)

        return super().add(
            key=key, value=value, version=version, client=client, timeout=timeout
        )

    def get_many(
        self,
        keys: KeyT,
        version: int | None = None,
        _client: Valkey | Any | None = None,
    ) -> OrderedDict:
        if not keys:
            return OrderedDict()

        recovered_data = OrderedDict()

        new_keys = [self.make_key(key, version=version) for key in keys]
        map_keys = dict(zip(new_keys, keys))

        for key in new_keys:
            client = self.get_client(key=key)
            value = self.get(key=key, version=version, client=client)

            if value is None:
                continue

            recovered_data[map_keys[key]] = value
        return recovered_data

    def mget(self, *args, **kwargs):
        raise NotImplementedError

    def set(
        self,
        key: KeyT,
        value: EncodableT,
        timeout: int | float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
        nx: bool = False,
        xx: bool = False,
    ) -> bool:
        """
        Persist a value to the cache, and set an optional expiration time.
        """
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_client(key=key)

        return super().set(
            key=key,
            value=value,
            timeout=timeout,
            version=version,
            client=client,
            nx=nx,
            xx=xx,
        )

    def set_many(
        self,
        data: Dict[KeyT, EncodableT],
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> None:
        """
        Set a bunch of values in the cache at once from a dict of key/value
        pairs. This is much more efficient than calling set() multiple times.

        If timeout is given, that timeout will be used for the key; otherwise
        the default cache timeout will be used.
        """
        for key, value in data.items():
            self.set(key, value, timeout, version=version, client=client)

    def mset(self, *args, **kwargs):
        raise NotImplementedError

    def delete_many(
        self, keys, version=None, _client: Valkey | Any | None = None
    ) -> int:
        """
        Remove multiple keys at once.
        """
        res = 0
        for key in [self.make_key(k, version=version) for k in keys]:
            client = self.get_client(key=key)
            res += self.delete(key, client=client)
        return res

    def incr_version(
        self,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_client(key=key)

        new_key, old_key, value, ttl, version = self._incr_version(
            key, delta, version, client
        )
        self.set(new_key, value, timeout=ttl, client=self.get_client(key=new_key))
        self.delete(old_key, client=client)
        return version + delta

    def iter_keys(
        self,
        search: str,
        itersize: int | None = None,
        client: Valkey | Any | None = None,
        version: int | None = None,
    ):
        """Not Implemented"""
        error_message = "iter_keys not supported on sharded client"
        raise NotImplementedError(error_message)

    def keys(
        self,
        search: str,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> List[str]:
        pattern = self.make_pattern(search, version=version)
        keys = []
        try:
            for connection in self._server_dict.values():
                keys.extend(connection.keys(pattern))
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._server_dict.values()) from e

        return [self.reverse_key(k.decode()) for k in keys]

    def delete_pattern(
        self,
        pattern: str,
        version: int | None = None,
        client: Valkey | Any | None = None,
        itersize: int | None = None,
        prefix: str | None = None,
    ) -> int:
        """
        Remove all keys matching pattern.
        """
        pattern = self.make_pattern(pattern, version=version, prefix=prefix)
        kwargs = {"match": pattern}
        if itersize:
            kwargs["count"] = itersize

        keys = []
        try:
            for connection in self._server_dict.values():
                keys.extend(key for key in connection.scan_iter(**kwargs))
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._server_dict.values()) from e

        res = 0
        if keys:
            try:
                for connection in self._server_dict.values():
                    res += connection.delete(*keys)
            except _main_exceptions as e:
                raise ConnectionInterrupted(
                    connection=self._server_dict.values()
                ) from e
        return res

    def _close(self) -> None:
        for client in self._server_dict.values():
            self.disconnect(client=client)

    def clear(self, client=None) -> None:
        for connection in self._server_dict.values():
            try:
                connection.flushdb()
            except _main_exceptions as e:
                raise ConnectionInterrupted(connection=connection) from e



import re
from contextlib import suppress
from typing import Any

from valkey.typing import KeyT, EncodableT

from django_valkey.compressors.base import BaseCompressor
from django_valkey.exceptions import CompressorError
from django_valkey.serializers.base import BaseSerializer


class CacheKey(str):
    """
    A stub string class that we can use to check if a key was created already.
    """

    def original_key(self) -> str:
        return self.rsplit(":", 1)[1]


def default_reverse_key(key: str) -> str:
    return key.split(":", 2)[2]


special_re = re.compile("([*?[])")


def glob_escape(s: str) -> str:
    return special_re.sub(r"[\1]", s)


def make_key(
    key: KeyT | None, key_func, version: int | None = None, prefix: str | None = None
) -> CacheKey | None:
    if not key:
        return key

    if isinstance(key, CacheKey):
        return key

    return CacheKey(key_func(key, prefix, version))


def make_pattern(
    pattern: str | None, key_func, version: int | None = None, prefix: str | None = None
) -> CacheKey | None:
    if not pattern:
        return pattern

    if isinstance(pattern, CacheKey):
        return pattern

    prefix = glob_escape(prefix)
    version_str = glob_escape(str(version))

    return CacheKey(key_func(pattern, prefix, version_str))


def encode(
    value: EncodableT, serializer: BaseSerializer, compressor: BaseCompressor
) -> bytes | int | float:
    if type(value) not in {int, float}:
        value = serializer.dumps(value)
        return compressor.compress(value)

    return value


def decode(value: bytes, serializer: BaseSerializer, compressor: BaseCompressor) -> Any:
    try:
        if value.isdigit():
            value = int(value)
        else:
            value = float(value)

    except (ValueError, TypeError):
        with suppress(CompressorError):
            value = compressor.decompress(value)
        value = serializer.loads(value)

    return value



SECRET_KEY = "django_tests_secret_key"

DJANGO_VALKEY_CONNECTION_FACTORY = (
    "django_valkey.async_cache.pool.AsyncSentinelConnectionFactory"
)

SENTINELS = [("127.0.0.1", 26379)]

CACHES = {
    "default": {
        "BACKEND": "django_valkey.async_cache.cache.AsyncValkeyCache",
        "LOCATION": ["valkey://mymaster?db=1"],
        "OPTIONS": {
            "CLIENT_CLASS": "django_valkey.async_cache.client.AsyncSentinelClient",
            "SENTINELS": SENTINELS,
        },
    },
    "doesnotexist": {
        "BACKEND": "django_valkey.async_cache.cache.AsyncValkeyCache",
        "LOCATION": "valkey://missing_service?db=1",
        "OPTIONS": {
            "CLIENT_CLASS": "django_valkey.async_cache.client.AsyncDefaultClient",
            "SENTINELS": SENTINELS,
        },
    },
    "sample": {
        "BACKEND": "django_valkey.async_cache.cache.AsyncValkeyCache",
        "LOCATION": "valkey://mymaster?db=1",
        "OPTIONS": {
            "CLIENT_CLASS": "django_valkey.async_cache.client.AsyncSentinelClient",
            "SENTINELS": SENTINELS,
        },
    },
    "with_prefix": {
        "BACKEND": "django_valkey.async_cache.cache.AsyncValkeyCache",
        "LOCATION": "valkey://mymaster?db=1",
        "KEY_PREFIX": "test-prefix",
        "OPTIONS": {
            "CLIENT_CLASS": "django_valkey.async_cache.client.AsyncDefaultClient",
            "SENTINELS": SENTINELS,
        },
    },
}

INSTALLED_APPS = ["django.contrib.sessions"]

USE_TZ = False

ROOT_URLCONF = "tests.settings.urls"



import datetime
import threading
import time
from collections.abc import Iterable
from datetime import timedelta
from typing import List, cast

import pytest
from pytest_django.fixtures import SettingsWrapper
from pytest_mock import MockerFixture

from django.core.cache import caches
from django.core.cache.backends.base import DEFAULT_TIMEOUT
from django.test import override_settings

from django_valkey.cache import ValkeyCache
from django_valkey.client import ShardClient, herd
from django_valkey.cluster_cache.client import DefaultClusterClient
from django_valkey.serializers.json import JSONSerializer
from django_valkey.serializers.msgpack import MSGPackSerializer
from django_valkey.serializers.msgspec import (
    MsgSpecJsonSerializer,
    MsgSpecMsgPackSerializer,
)
from django_valkey.serializers.pickle import PickleSerializer


@pytest.fixture
def patch_itersize_setting() -> Iterable[None]:
    # destroy cache to force recreation with overridden settings
    del caches["default"]
    with override_settings(DJANGO_VALKEY_SCAN_ITERSIZE=30):
        yield
    # destroy cache to force recreation with original settings
    del caches["default"]


class TestDjangoValkeyCache:
    def test_set_int(self, cache: ValkeyCache):
        if isinstance(cache.client, herd.HerdClient):
            pytest.skip("herd client's set method works differently")
        cache.set("test_key", 1)
        result = cache.get("test_key")
        assert type(result) is int
        # shard client doesn't have get_client()
        if not isinstance(cache.client, ShardClient):
            raw_client = cache.client._get_client(write=False, client=None)
        else:
            raw_client = cache.client._get_client(key=":1:test_key")
        assert raw_client.get(":1:test_key") == b"1"

    def test_set_float(self, cache: ValkeyCache):
        if isinstance(cache.client, herd.HerdClient):
            pytest.skip("herd client's set method works differently")
        cache.set("test_key2", 1.1)
        result = cache.get("test_key2")
        assert type(result) is float
        if not isinstance(cache.client, ShardClient):
            raw_client = cache.client._get_client(write=False, client=None)
        else:
            raw_client = cache.client._get_client(key=":1:test_key2")
        assert raw_client.get(":1:test_key2") == b"1.1"

    def test_setnx(self, cache: ValkeyCache):
        # we should ensure there is no test_key_nx in valkey
        cache.delete("test_key_nx")
        res = cache.get("test_key_nx")
        assert res is None

        res = cache.set("test_key_nx", 1, nx=True)
        assert bool(res) is True
        # test that second set will have
        res = cache.set("test_key_nx", 2, nx=True)
        assert res is None
        res = cache.get("test_key_nx")
        assert res == 1

        cache.delete("test_key_nx")
        res = cache.get("test_key_nx")
        assert res is None

    def test_setnx_timeout(self, cache: ValkeyCache):
        # test that timeout still works for nx=True
        res = cache.set("test_key_nx", 1, timeout=2, nx=True)
        assert res is True
        time.sleep(3)
        res = cache.get("test_key_nx")
        assert res is None

        # test that timeout will not affect key, if it was there
        cache.set("test_key_nx", 1)
        res = cache.set("test_key_nx", 2, timeout=2, nx=True)
        assert res is None
        time.sleep(3)
        res = cache.get("test_key_nx")
        assert res == 1

        cache.delete("test_key_nx")
        res = cache.get("test_key_nx")
        assert res is None

    def test_unicode_keys(self, cache: ValkeyCache):
        cache.set("ключ", "value")
        res = cache.get("ключ")
        assert res == "value"

    def test_save_an_integer(self, cache: ValkeyCache):
        cache.set("test_key", 2)
        res = cache.get("test_key", "Foo")

        assert isinstance(res, int)
        assert res == 2

    def test_save_string(self, cache: ValkeyCache):
        cache.set("test_key", "hello" * 1000)
        res = cache.get("test_key")

        assert isinstance(res, str)
        assert res == "hello" * 1000

        cache.set("test_key", "2")
        res = cache.get("test_key")

        assert isinstance(res, str)
        assert res == "2"

    def test_save_unicode(self, cache: ValkeyCache):
        cache.set("test_key", "heló")
        res = cache.get("test_key")

        assert isinstance(res, str)
        assert res == "heló"

    def test_save_dict(self, cache: ValkeyCache):
        if isinstance(
            cache.client._serializer,
            (
                JSONSerializer,
                MSGPackSerializer,
                MsgSpecJsonSerializer,
                MsgSpecMsgPackSerializer,
            ),
        ):
            # JSONSerializer and MSGPackSerializer use the isoformat for
            # datetimes.
            now_dt: str | datetime.datetime = datetime.datetime.now().isoformat()
        else:
            now_dt = datetime.datetime.now()

        test_dict = {"id": 1, "date": now_dt, "name": "Foo"}

        cache.set("test_key", test_dict)
        res = cache.get("test_key")

        assert isinstance(res, dict)
        assert res["id"] == 1
        assert res["name"] == "Foo"
        assert res["date"] == now_dt

    def test_save_float(self, cache: ValkeyCache):
        float_val = 1.345620002

        cache.set("test_key", float_val)
        res = cache.get("test_key")

        assert isinstance(res, float)
        assert res == float_val

    def test_timeout(self, cache: ValkeyCache):
        cache.set("test_key", 222, timeout=3)
        time.sleep(4)

        res = cache.get("test_key")
        assert res is None

    def test_timeout_0(self, cache: ValkeyCache):
        cache.set("test_key", 222, timeout=0)
        res = cache.get("test_key")
        assert res is None

    def test_timeout_parameter_as_positional_argument(self, cache: ValkeyCache):
        cache.set("test_key", 222, -1)
        res = cache.get("test_key")
        assert res is None

        cache.set("test_key", 222, 1)
        res1 = cache.get("test_key")
        time.sleep(2)
        res2 = cache.get("test_key")
        assert res1 == 222
        assert res2 is None

        # nx=True should not overwrite expire of key already in db
        cache.set("test_key", 222, None)
        cache.set("test_key", 222, -1, nx=True)
        res = cache.get("test_key")
        assert res == 222

    def test_timeout_negative(self, cache: ValkeyCache):
        cache.set("test_key", 222, timeout=-1)
        res = cache.get("test_key")
        assert res is None

        cache.set("test_key", 222, timeout=None)
        cache.set("test_key", 222, timeout=-1)
        res = cache.get("test_key")
        assert res is None

        # nx=True should not overwrite expire of key already in db
        cache.set("test_key", 222, timeout=None)
        cache.set("test_key", 222, timeout=-1, nx=True)
        res = cache.get("test_key")
        assert res == 222

    def test_timeout_tiny(self, cache: ValkeyCache):
        cache.set("test_key", 222, timeout=0.00001)
        res = cache.get("test_key")
        assert res in (None, 222)

    def test_set_add(self, cache: ValkeyCache):
        cache.set("add_key", "Initial value")
        res = cache.add("add_key", "New value")
        assert res is None

        res = cache.get("add_key")
        assert res == "Initial value"
        res = cache.add("other_key", "New value")
        assert res is True

    def test_get_many(self, cache: ValkeyCache):
        cache.set("a", 1)
        cache.set("b", 2)
        cache.set("c", 3)

        res = cache.get_many(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    def test_mget(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient) or isinstance(
            cache.client, DefaultClusterClient
        ):
            pytest.skip()
        cache.set("a", 1)
        cache.set("b", 2)
        cache.set("c", 3)

        res = cache.mget(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    def test_get_many_unicode(self, cache: ValkeyCache):
        cache.set("a", "1")
        cache.set("ب", "2")
        cache.set("c", "الف")

        res = cache.get_many(["a", "ب", "c"])
        assert res == {"a": "1", "ب": "2", "c": "الف"}

    def test_mget_unicode(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient) or isinstance(
            cache.client, DefaultClusterClient
        ):
            pytest.skip()

        cache.set("fooa", "1")
        cache.set("fooب", "2")
        cache.set("fooc", "الف")

        res = cache.mget(["fooa", "fooب", "fooc"])
        assert res == {"fooa": "1", "fooب": "2", "fooc": "الف"}

    def test_set_many(self, cache: ValkeyCache):
        cache.set_many({"a": 1, "b": 2, "c": 3})
        res = cache.get_many(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    def test_mset(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient) or isinstance(
            cache.client, DefaultClusterClient
        ):
            pytest.skip()
        cache.mset({"a": 1, "b": 2, "c": 3})
        res = cache.mget(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    def test_set_call_empty_pipeline(
        self,
        cache: ValkeyCache,
        mocker: MockerFixture,
        settings: SettingsWrapper,
    ):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support get_client")

        pipeline = cache.client.get_client(write=True).pipeline()
        key = "key"
        value = "value"

        mocked_set = mocker.patch.object(pipeline, "set")

        cache.set(key, value, client=pipeline)

        if isinstance(cache.client, herd.HerdClient):
            default_timeout = cache.client._backend.default_timeout
            herd_timeout = (default_timeout + settings.CACHE_HERD_TIMEOUT) * 1000
            herd_pack_value = cache.client._pack(value, default_timeout)
            mocked_set.assert_called_once_with(
                cache.client.make_key(key, version=None),
                cache.client.encode(herd_pack_value),
                nx=False,
                px=herd_timeout,
                xx=False,
            )
        else:
            mocked_set.assert_called_once_with(
                cache.client.make_key(key, version=None),
                cache.client.encode(value),
                nx=False,
                px=cache.client._backend.default_timeout * 1000,
                xx=False,
            )

    def test_delete(self, cache: ValkeyCache):
        cache.set_many({"a": 1, "b": 2, "c": 3})
        res = cache.delete("a")
        assert bool(res) is True

        res = cache.get_many(["a", "b", "c"])
        assert res == {"b": 2, "c": 3}

        res = cache.delete("a")
        assert bool(res) is False

    def test_delete_return_value_type(self, cache: ValkeyCache):
        """delete() returns a boolean instead of int since django version 3.1"""
        cache.set("a", 1)
        res = cache.delete("a")
        assert isinstance(res, bool)
        assert res is True
        res = cache.delete("b")
        assert isinstance(res, bool)
        assert res is False

    def test_delete_many(self, cache: ValkeyCache):
        cache.set_many({"a": 1, "b": 2, "c": 3})
        res = cache.delete_many(["a", "b"])
        assert bool(res) is True

        res = cache.get_many(["a", "b", "c"])
        assert res == {"c": 3}

        res = cache.delete_many(["a", "b"])
        assert bool(res) is False

    def test_delete_many_generator(self, cache: ValkeyCache):
        cache.set_many({"a": 1, "b": 2, "c": 3})
        res = cache.delete_many(key for key in ["a", "b"])
        assert bool(res) is True

        res = cache.get_many(["a", "b", "c"])
        assert res == {"c": 3}

        res = cache.delete_many(["a", "b"])
        assert bool(res) is False

    def test_delete_many_empty_generator(self, cache: ValkeyCache):
        res = cache.delete_many(key for key in cast(List[str], []))
        assert bool(res) is False

    def test_incr(self, cache: ValkeyCache):
        if isinstance(cache.client, herd.HerdClient):
            pytest.skip("HerdClient doesn't support incr")

        cache.set("num", 1)

        cache.incr("num")
        res = cache.get("num")
        assert res == 2

        cache.incr("num", 10)
        res = cache.get("num")
        assert res == 12

        # max 64-bit signed int
        cache.set("num", 9223372036854775807)

        cache.incr("num")
        res = cache.get("num")
        assert res == 9223372036854775808

        cache.incr("num", 2)
        res = cache.get("num")
        assert res == 9223372036854775810

        cache.set("num", 3)

        cache.incr("num", 2)
        res = cache.get("num")
        assert res == 5

    def test_incr_no_timeout(self, cache: ValkeyCache):
        if isinstance(cache.client, herd.HerdClient):
            pytest.skip("HerdClient doesn't support incr")

        cache.set("num", 1, timeout=None)

        cache.incr("num")
        res = cache.get("num")
        assert res == 2

        cache.incr("num", 10)
        res = cache.get("num")
        assert res == 12

        # max 64-bit signed int
        cache.set("num", 9223372036854775807, timeout=None)

        cache.incr("num")
        res = cache.get("num")
        assert res == 9223372036854775808

        cache.incr("num", 2)
        res = cache.get("num")
        assert res == 9223372036854775810

        cache.set("num", 3, timeout=None)

        cache.incr("num", 2)
        res = cache.get("num")
        assert res == 5

    def test_incr_error(self, cache: ValkeyCache):
        if isinstance(cache.client, herd.HerdClient):
            pytest.skip("HerdClient doesn't support incr")

        with pytest.raises(ValueError):
            # key does not exist
            cache.incr("numnum")

    def test_incr_ignore_check(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support argument ignore_key_check to incr")
        if isinstance(cache.client, herd.HerdClient):
            pytest.skip("HerdClient doesn't support incr")

        # key exists check will be skipped and the value will be incremented by
        # '1' which is the default delta
        cache.incr("num", ignore_key_check=True)
        res = cache.get("num")
        assert res == 1
        cache.delete("num")

        # since key doesn't exist it is set to the delta value, 10 in this case
        cache.incr("num", 10, ignore_key_check=True)
        res = cache.get("num")
        assert res == 10
        cache.delete("num")

        # following are just regression checks to make sure it still works as
        # expected with incr max 64-bit signed int
        cache.set("num", 9223372036854775807)

        cache.incr("num", ignore_key_check=True)
        res = cache.get("num")
        assert res == 9223372036854775808

        cache.incr("num", 2, ignore_key_check=True)
        res = cache.get("num")
        assert res == 9223372036854775810

        cache.set("num", 3)

        cache.incr("num", 2, ignore_key_check=True)
        res = cache.get("num")
        assert res == 5

    def test_get_set_bool(self, cache: ValkeyCache):
        cache.set("bool", True)
        res = cache.get("bool")

        assert isinstance(res, bool)
        assert res is True

        cache.set("bool", False)
        res = cache.get("bool")

        assert isinstance(res, bool)
        assert res is False

    def test_decr(self, cache: ValkeyCache):
        if isinstance(cache.client, herd.HerdClient):
            pytest.skip("HerdClient doesn't support decr")

        cache.set("num", 20)

        cache.decr("num")
        res = cache.get("num")
        assert res == 19

        cache.decr("num", 20)
        res = cache.get("num")
        assert res == -1

        cache.decr("num", 2)
        res = cache.get("num")
        assert res == -3

        cache.set("num", 20)

        cache.decr("num")
        res = cache.get("num")
        assert res == 19

        # max 64-bit signed int + 1
        cache.set("num", 9223372036854775808)

        cache.decr("num")
        res = cache.get("num")
        assert res == 9223372036854775807

        cache.decr("num", 2)
        res = cache.get("num")
        assert res == 9223372036854775805

    def test_version(self, cache: ValkeyCache):
        cache.set("keytest", 2, version=2)
        res = cache.get("keytest")
        assert res is None

        res = cache.get("keytest", version=2)
        assert res == 2

    def test_incr_version(self, cache: ValkeyCache):
        cache.set("keytest", 2)
        cache.incr_version("keytest")

        res = cache.get("keytest")
        assert res is None

        res = cache.get("keytest", version=2)
        assert res == 2

    def test_ttl_incr_version_no_timeout(self, cache: ValkeyCache):
        cache.set("my_key", "hello world!", timeout=None)

        cache.incr_version("my_key")

        my_value = cache.get("my_key", version=2)

        assert my_value == "hello world!"

    def test_decr_version(self, cache: ValkeyCache):
        cache.set("keytest", 2, version=3)
        res = cache.get("keytest", version=3)
        assert res == 2

        cache.decr_version("keytest", version=3)

        res = cache.get("keytest", version=3)
        assert res is None

        res = cache.get("keytest", version=2)
        assert res == 2

    def test_ttl_decr_version_no_timeout(self, cache: ValkeyCache):
        cache.set("my_key", "hello world!", version=3, timeout=None)

        cache.decr_version("my_key", version=3)

        my_value = cache.get("my_key", version=2)

        assert my_value == "hello world!"

    def test_delete_pattern(self, cache: ValkeyCache):
        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
            cache.set(key, "foo")

        res = cache.delete_pattern("*foo-a*")
        assert bool(res) is True

        keys = cache.keys("foo*")
        assert set(keys) == {"foo-bb", "foo-bc"}

        res = cache.delete_pattern("*foo-a*")
        assert bool(res) is False

    def test_delete_pattern_with_custom_count(self, cache: ValkeyCache, mocker):
        client_mock = mocker.patch("django_valkey.cache.ValkeyCache.client")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
            cache.set(key, "foo")

        cache.delete_pattern("*foo-a*", itersize=2)

        client_mock.delete_pattern.assert_called_once_with("*foo-a*", itersize=2)

    def test_delete_pattern_with_settings_default_scan_count(
        self,
        patch_itersize_setting,
        cache: ValkeyCache,
        settings: SettingsWrapper,
        mocker,
    ):
        client_mock = mocker.patch("django_valkey.cache.ValkeyCache.client")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
            cache.set(key, "foo")
        expected_count = settings.DJANGO_VALKEY_SCAN_ITERSIZE

        cache.delete_pattern("*foo-a*")

        client_mock.delete_pattern.assert_called_once_with(
            "*foo-a*", itersize=expected_count
        )

    def test_close(self, cache: ValkeyCache, settings: SettingsWrapper):
        settings.DJANGO_VALKEY_CLOSE_CONNECTION = True
        cache.set("f", "1")
        cache.close()

    def test_close_client(self, cache: ValkeyCache, mocker: MockerFixture):
        mock = mocker.patch.object(cache.client, "close")

        cache.close()
        assert mock.called

    def test_ttl(self, cache: ValkeyCache):
        cache.set("foo", "bar", 10)
        ttl = cache.ttl("foo")

        if isinstance(cache.client, herd.HerdClient):
            assert pytest.approx(ttl) == 12
        else:
            assert pytest.approx(ttl) == 10

        # Test ttl None
        cache.set("foo", "foo", timeout=None)
        ttl = cache.ttl("foo")
        assert ttl is None

        # Test ttl with expired key
        cache.set("foo", "foo", timeout=-1)
        ttl = cache.ttl("foo")
        assert ttl == 0

        # Test ttl with not existent key
        ttl = cache.ttl("not-existent-key")
        assert ttl == 0

    def test_pttl(self, cache: ValkeyCache):
        # Test pttl
        cache.set("foo", "bar", 10)
        ttl = cache.pttl("foo")

        # delta is set to 10 as precision error causes tests to fail
        if isinstance(cache.client, herd.HerdClient):
            assert pytest.approx(ttl, 10) == 12000
        else:
            assert pytest.approx(ttl, 10) == 10000

        # Test pttl with float value
        cache.set("foo", "bar", 5.5)
        ttl = cache.pttl("foo")

        if isinstance(cache.client, herd.HerdClient):
            assert pytest.approx(ttl, 10) == 7500
        else:
            assert pytest.approx(ttl, 10) == 5500

        # Test pttl None
        cache.set("foo", "foo", timeout=None)
        ttl = cache.pttl("foo")
        assert ttl is None

        # Test pttl with expired key
        cache.set("foo", "foo", timeout=-1)
        ttl = cache.pttl("foo")
        assert ttl == 0

        # Test pttl with not existent key
        ttl = cache.pttl("not-existent-key")
        assert ttl == 0

    def test_persist(self, cache: ValkeyCache):
        cache.set("foo", "bar", timeout=20)
        assert cache.persist("foo") is True

        ttl = cache.ttl("foo")
        assert ttl is None
        assert cache.persist("not-existent-key") is False

    def test_expire(self, cache: ValkeyCache):
        cache.set("foo", "bar", timeout=None)
        assert cache.expire("foo", 20) is True
        ttl = cache.ttl("foo")
        assert pytest.approx(ttl) == 20
        assert cache.expire("not-existent-key", 20) is False

    def test_expire_with_default_timeout(self, cache: ValkeyCache):
        cache.set("foo", "bar", timeout=None)
        assert cache.expire("foo", DEFAULT_TIMEOUT) is True
        assert cache.expire("not-existent-key", DEFAULT_TIMEOUT) is False

    def test_pexpire(self, cache: ValkeyCache):
        cache.set("foo", "bar", timeout=None)
        assert cache.pexpire("foo", 20500) is True
        ttl = cache.pttl("foo")
        # delta is set to 10 as precision error causes tests to fail
        assert pytest.approx(ttl, 10) == 20500
        assert cache.pexpire("not-existent-key", 20500) is False

    def test_pexpire_with_default_timeout(self, cache: ValkeyCache):
        cache.set("foo", "bar", timeout=None)
        assert cache.pexpire("foo", DEFAULT_TIMEOUT) is True
        assert cache.pexpire("not-existent-key", DEFAULT_TIMEOUT) is False

    def test_pexpire_at(self, cache: ValkeyCache):
        # Test settings expiration time 1 hour ahead by datetime.
        cache.set("foo", "bar", timeout=None)
        expiration_time = datetime.datetime.now() + timedelta(hours=1)
        assert cache.pexpire_at("foo", expiration_time) is True
        ttl = cache.pttl("foo")
        assert pytest.approx(ttl, 10) == timedelta(hours=1).total_seconds()

        # Test settings expiration time 1 hour ahead by Unix timestamp.
        cache.set("foo", "bar", timeout=None)
        expiration_time = datetime.datetime.now() + timedelta(hours=2)
        assert cache.pexpire_at("foo", int(expiration_time.timestamp() * 1000)) is True
        ttl = cache.pttl("foo")
        assert pytest.approx(ttl, 10) == timedelta(hours=2).total_seconds() * 1000

        # Test settings expiration time 1 hour in the past, which effectively
        # deletes the key.
        expiration_time = datetime.datetime.now() - timedelta(hours=2)
        assert cache.pexpire_at("foo", expiration_time) is True
        value = cache.get("foo")
        assert value is None

        expiration_time = datetime.datetime.now() + timedelta(hours=2)
        assert cache.pexpire_at("not-existent-key", expiration_time) is False

    def test_expire_at(self, cache: ValkeyCache):
        # Test settings expiration time 1 hour ahead by datetime.
        cache.set("foo", "bar", timeout=None)
        expiration_time = datetime.datetime.now() + timedelta(hours=1)
        assert cache.expire_at("foo", expiration_time) is True
        ttl = cache.ttl("foo")
        assert pytest.approx(ttl, 1) == timedelta(hours=1).total_seconds()

        # Test settings expiration time 1 hour ahead by Unix timestamp.
        cache.set("foo", "bar", timeout=None)
        expiration_time = datetime.datetime.now() + timedelta(hours=2)
        assert cache.expire_at("foo", int(expiration_time.timestamp())) is True
        ttl = cache.ttl("foo")
        assert pytest.approx(ttl, 1) == timedelta(hours=1).total_seconds() * 2

        # Test settings expiration time 1 hour in the past, which effectively
        # deletes the key.
        expiration_time = datetime.datetime.now() - timedelta(hours=2)
        assert cache.expire_at("foo", expiration_time) is True
        value = cache.get("foo")
        assert value is None

        expiration_time = datetime.datetime.now() + timedelta(hours=2)
        assert cache.expire_at("not-existent-key", expiration_time) is False

    def test_lock(self, cache: ValkeyCache):
        lock = cache.lock("foobar")
        lock.acquire(blocking=True)

        assert cache.has_key("foobar")
        lock.release()
        assert not cache.has_key("foobar")

    def test_lock_context_manager(self, cache: ValkeyCache):
        with cache.lock("foobar"):
            assert cache.has_key("foobar")
        assert not cache.has_key("foobar")

    def test_lock_released_by_thread(self, cache: ValkeyCache):
        lock = cache.lock("foobar", thread_local=False)
        lock.acquire(blocking=True)

        def release_lock(lock_):
            lock_.release()

        t = threading.Thread(target=release_lock, args=[lock])
        t.start()
        t.join()

        assert not cache.has_key("foobar")

    def test_iter_keys(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support iter_keys")

        cache.set("foo1", 1)
        cache.set("foo2", 1)
        cache.set("foo3", 1)

        # Test simple result
        result = set(cache.iter_keys("foo*"))
        assert result == {"foo1", "foo2", "foo3"}

    def test_iter_keys_itersize(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support iter_keys")

        cache.set("foo1", 1)
        cache.set("foo2", 1)
        cache.set("foo3", 1)

        # Test limited result
        result = list(cache.iter_keys("foo*", itersize=2))
        assert len(result) == 3

    def test_iter_keys_generator(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support iter_keys")

        cache.set("foo1", 1)
        cache.set("foo2", 1)
        cache.set("foo3", 1)

        # Test generator object
        result = cache.iter_keys("foo*")
        next_value = next(result)
        assert next_value is not None

    def test_primary_replica_switching(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("shard client handles connections differently")
        cache = cast(ValkeyCache, caches["sample"])
        client = cache.client
        client._server = ["foo", "bar"]
        client._clients = ["Foo", "Bar"]

        assert client.get_client(write=True) == "Foo"
        assert client.get_client(write=False) == "Bar"

    def test_primary_replica_switching_with_index(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support get_client")

        cache = cast(ValkeyCache, caches["sample"])
        client = cache.client
        client._server = ["foo", "bar"]
        client._clients = ["Foo", "Bar"]

        assert client.get_client_with_index(write=True) == ("Foo", 0)
        assert client.get_client_with_index(write=False) == ("Bar", 1)

    def test_touch_zero_timeout(self, cache: ValkeyCache):
        cache.set("test_key", 222, timeout=10)

        assert cache.touch("test_key", 0) is True
        res = cache.get("test_key")
        assert res is None

    def test_touch_positive_timeout(self, cache: ValkeyCache):
        cache.set("test_key", 222, timeout=10)

        assert cache.touch("test_key", 2) is True
        assert cache.get("test_key") == 222
        time.sleep(3)
        assert cache.get("test_key") is None

    def test_touch_negative_timeout(self, cache: ValkeyCache):
        cache.set("test_key", 222, timeout=10)

        assert cache.touch("test_key", -1) is True
        res = cache.get("test_key")
        assert res is None

    def test_touch_missed_key(self, cache: ValkeyCache):
        assert cache.touch("test_key_does_not_exist", 1) is False

    def test_touch_forever(self, cache: ValkeyCache):
        cache.set("test_key", "foo", timeout=1)
        result = cache.touch("test_key", None)
        assert result is True
        assert cache.ttl("test_key") is None
        time.sleep(2)
        assert cache.get("test_key") == "foo"

    def test_touch_forever_nonexistent(self, cache: ValkeyCache):
        result = cache.touch("test_key_does_not_exist", None)
        assert result is False

    def test_touch_default_timeout(self, cache: ValkeyCache):
        cache.set("test_key", "foo", timeout=1)
        result = cache.touch("test_key")
        assert result is True
        time.sleep(2)
        assert cache.get("test_key") == "foo"

    def test_clear(self, cache: ValkeyCache):
        cache.set("foo", "bar")
        value_from_cache = cache.get("foo")
        assert value_from_cache == "bar"
        cache.clear()
        value_from_cache_after_clear = cache.get("foo")
        assert value_from_cache_after_clear is None

    def test_hset(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support hash operations")
        cache.hset("foo_hash1", "foo1", "bar1")
        cache.hset("foo_hash1", "foo2", "bar2")
        assert cache.hlen("foo_hash1") == 2
        assert cache.hexists("foo_hash1", "foo1")
        assert cache.hexists("foo_hash1", "foo2")

    def test_hdel(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support hash operations")
        cache.hset("foo_hash2", "foo1", "bar1")
        cache.hset("foo_hash2", "foo2", "bar2")
        assert cache.hlen("foo_hash2") == 2
        deleted_count = cache.hdel("foo_hash2", "foo1")
        assert deleted_count == 1
        assert cache.hlen("foo_hash2") == 1
        assert not cache.hexists("foo_hash2", "foo1")
        assert cache.hexists("foo_hash2", "foo2")

    def test_hlen(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support hash operations")
        assert cache.hlen("foo_hash3") == 0
        cache.hset("foo_hash3", "foo1", "bar1")
        assert cache.hlen("foo_hash3") == 1
        cache.hset("foo_hash3", "foo2", "bar2")
        assert cache.hlen("foo_hash3") == 2

    def test_hkeys(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support hash operations")
        cache.hset("foo_hash4", "foo1", "bar1")
        cache.hset("foo_hash4", "foo2", "bar2")
        cache.hset("foo_hash4", "foo3", "bar3")
        keys = cache.hkeys("foo_hash4")
        assert len(keys) == 3
        for i in range(len(keys)):
            assert keys[i] == f"foo{i + 1}"

    def test_hexists(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support hash operations")
        cache.hset("foo_hash5", "foo1", "bar1")
        assert cache.hexists("foo_hash5", "foo1")
        assert not cache.hexists("foo_hash5", "foo")

    def test_sadd(self, cache: ValkeyCache):
        assert cache.sadd("foo", "bar") == 1
        assert cache.smembers("foo") == {"bar"}

    def test_sadd_int(self, cache: ValkeyCache):
        cache.sadd("foo", 1)
        assert cache.smembers("foo") == {1}
        if not isinstance(cache.client, ShardClient):
            raw_client = cache.client._get_client(write=False, client=None)
        else:
            raw_client = cache.client._get_client(key=":1:foo")
        assert raw_client.smembers(":1:foo") == [b"1"]

    def test_sadd_float(self, cache: ValkeyCache):
        cache.sadd("foo", 1.2)
        assert cache.smembers("foo") == {1.2}
        if not isinstance(cache.client, ShardClient):
            raw_client = cache.client._get_client(write=False, client=None)
        else:
            raw_client = cache.client._get_client(key=":1:foo")
        assert raw_client.smembers(":1:foo") == [b"1.2"]

    def test_scard(self, cache: ValkeyCache):
        cache.sadd("foo", "bar", "bar2")
        assert cache.scard("foo") == 2

    def test_sdiff(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support sdiff")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2")
        cache.sadd("foo2", "bar2", "bar3")
        assert cache.sdiff("foo1", "foo2") == {"bar1"}

    def test_sdiffstore(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support sdiffstore")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2")
        cache.sadd("foo2", "bar2", "bar3")
        assert cache.sdiffstore("foo3", "foo1", "foo2") == 1
        assert cache.smembers("foo3") == {"bar1"}

    def test_sdiffstore_with_keys_version(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support sdiffstore")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2", version=2)
        cache.sadd("foo2", "bar2", "bar3", version=2)
        assert cache.sdiffstore("foo3", "foo1", "foo2", version_keys=2) == 1
        assert cache.smembers("foo3") == {"bar1"}

    def test_sdiffstore_with_different_keys_versions_without_initial_set_in_version(
        self, cache: ValkeyCache
    ):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support sdiffstore")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2", version=1)
        cache.sadd("foo2", "bar2", "bar3", version=2)
        assert cache.sdiffstore("foo3", "foo1", "foo2", version_keys=2) == 0

    def test_sdiffstore_with_different_keys_versions_with_initial_set_in_version(
        self, cache: ValkeyCache
    ):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support sdiffstore")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2", version=2)
        cache.sadd("foo2", "bar2", "bar3", version=1)
        assert cache.sdiffstore("foo3", "foo1", "foo2", version_keys=2) == 2

    def test_sinter(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support sinter")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2")
        cache.sadd("foo2", "bar2", "bar3")
        assert cache.sinter("foo1", "foo2") == {"bar2"}

    def test_sinterstore(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support sinterstore")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2")
        cache.sadd("foo2", "bar2", "bar3")
        assert cache.sinterstore("foo3", "foo1", "foo2") == 1
        assert cache.smembers("foo3") == {"bar2"}

    def test_sismember_str(self, cache: ValkeyCache):
        cache.sadd("foo", "bar")
        assert cache.sismember("foo", "bar") is True
        assert cache.sismember("foo", "bar2") is False

    def test_sismember_int(self, cache: ValkeyCache):
        cache.sadd("baz", 3)
        assert cache.sismember("baz", 3) is True
        assert cache.sismember("baz", 2) is False

    def test_sismember_float(self, cache: ValkeyCache):
        cache.sadd("foo", 3.0)
        assert cache.sismember("foo", 3.0) is True
        assert cache.sismember("foo", 2.0) is False

    def test_sismember_byte(self, cache: ValkeyCache):
        if isinstance(cache.client._serializer, JSONSerializer):
            pytest.skip("JSONSerializer doesn't support the byte type")
        cache.sadd("foo", b"abc")
        assert cache.sismember("foo", b"abc") is True
        assert cache.sismember("foo", b"def") is False

    def test_sismember_bytearray(self, cache: ValkeyCache):
        if isinstance(cache.client._serializer, JSONSerializer):
            pytest.skip("JSONSerializer doesn't support the bytearray type")
        right_val = bytearray(b"abc")
        wrong_val = bytearray(b"def")
        cache.sadd("foo", right_val)
        assert cache.sismember("foo", right_val) is True
        assert cache.sismember("foo", wrong_val) is False

    def test_sismember_memoryview(self, cache: ValkeyCache):
        if isinstance(cache.client._serializer, (PickleSerializer, JSONSerializer)):
            pytest.skip(
                "PickleSerializer/JSONSerializer doesn't support the memoryview type"
            )
        right_val = memoryview(b"abc")
        wrong_val = memoryview(b"def")
        cache.sadd("foo", right_val)
        assert cache.sismember("foo", right_val) is True
        assert cache.sismember("foo", wrong_val) is False

    def test_sismember_complex(self, cache: ValkeyCache):
        if isinstance(
            cache.client._serializer,
            (
                JSONSerializer,
                MSGPackSerializer,
                MsgSpecJsonSerializer,
                MsgSpecMsgPackSerializer,
            ),
        ):
            pytest.skip(
                "JSONSerializer/MSGPackSerializer and neither the msgspec serializers don't support the complex type"
            )
        cache.sadd("foo", 3j)
        assert cache.sismember("foo", 3j) is True
        assert cache.sismember("foo", 4j) is False

    def test_sismember_list(self, cache: ValkeyCache):
        cache.sadd("foo", [1, 2, 3])
        assert cache.sismember("foo", [1, 2, 3]) is True
        assert cache.sismember("foo", [1, 2, 4]) is False

    def test_sismember_tuple(self, cache: ValkeyCache):
        cache.sadd("foo", (1, 2, 3))
        assert cache.sismember("foo", (1, 2, 3)) is True
        assert cache.sismember("foo", (1, 2, 4)) is False

    def test_sismember_set(self, cache: ValkeyCache):
        if isinstance(cache.client._serializer, (MSGPackSerializer, JSONSerializer)):
            pytest.skip("MSGPackSerializer doesn't support the set type")
        cache.sadd("foo", {1, 2, 3})
        assert cache.sismember("foo", {1, 2, 3}) is True
        assert cache.sismember("foo", {1, 2, 4}) is False

    def test_sismember_frozenset(self, cache: ValkeyCache):
        if isinstance(cache.client._serializer, (MSGPackSerializer, JSONSerializer)):
            pytest.skip("MSGPackSerializer doesn't support the frozenset type")
        cache.sadd("foo", frozenset(("a", "b")))
        assert cache.sismember("foo", frozenset(("a", "b"))) is True
        assert cache.sismember("foo", frozenset(("d", "c"))) is False

    def test_sismember_dict(self, cache: ValkeyCache):
        cache.sadd("foo", {"a": 1, "b": 2})
        assert cache.sismember("foo", {"a": 1, "b": 2}) is True
        assert cache.sismember("foo", {"a": 1, "c": 3}) is False

    def test_sismember_bool(self, cache: ValkeyCache):
        cache.sadd("foo", True)
        assert cache.sismember("foo", True) is True
        assert cache.sismember("foo", False) is False

    def test_smove(self, cache: ValkeyCache):
        # if isinstance(cache.client, ShardClient):
        #     pytest.skip("ShardClient doesn't support get_client")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2")
        cache.sadd("foo2", "bar2", "bar3")
        assert cache.smove("foo1", "foo2", "bar1") is True
        assert cache.smove("foo1", "foo2", "bar4") is False
        assert cache.smembers("foo1") == {"bar2"}
        assert cache.smembers("foo2") == {"bar1", "bar2", "bar3"}

    def test_spop_default_count(self, cache: ValkeyCache):
        cache.sadd("foo", "bar1", "bar2")
        assert cache.spop("foo") in {"bar1", "bar2"}
        assert cache.smembers("foo") in [{"bar1"}, {"bar2"}]

    def test_spop(self, cache: ValkeyCache):
        cache.sadd("foo", "bar1", "bar2")
        assert cache.spop("foo", 1) in [{"bar1"}, {"bar2"}]
        assert cache.smembers("foo") in [{"bar1"}, {"bar2"}]

    def test_srandmember_default_count(self, cache: ValkeyCache):
        cache.sadd("foo", "bar1", "bar2")
        assert cache.srandmember("foo") in {"bar1", "bar2"}

    def test_srandmember(self, cache: ValkeyCache):
        cache.sadd("foo", "bar1", "bar2")
        assert cache.srandmember("foo", 1) in [["bar1"], ["bar2"]]

    def test_srem(self, cache: ValkeyCache):
        cache.sadd("foo", "bar1", "bar2")
        assert cache.srem("foo", "bar1") == 1
        assert cache.srem("foo", "bar3") == 0

    def test_sscan(self, cache: ValkeyCache):
        cache.sadd("foo", "bar1", "bar2")
        items = cache.sscan("foo")
        assert items == {"bar1", "bar2"}

    def test_sscan_with_match(self, cache: ValkeyCache):
        if cache.client._has_compression_enabled():
            pytest.skip("Compression is enabled, sscan with match is not supported")
        cache.sadd("foo", "bar1", "bar2", "zoo")
        items = cache.sscan("foo", match="zoo")
        assert items == {"zoo"}

    def test_sscan_iter(self, cache: ValkeyCache):
        cache.sadd("foo", "bar1", "bar2")
        items = cache.sscan_iter("foo")
        assert set(items) == {"bar1", "bar2"}

    def test_sscan_iter_with_match(self, cache: ValkeyCache):
        if cache.client._has_compression_enabled():
            pytest.skip(
                "Compression is enabled, sscan_iter with match is not supported"
            )
        cache.sadd("foo", "bar1", "bar2", "zoo")
        items = cache.sscan_iter("foo", match="bar*")
        assert set(items) == {"bar1", "bar2"}

    def test_smismember(self, cache: ValkeyCache):
        cache.sadd("foo", "bar1", "bar2", "bar3")
        assert cache.smismember("foo", "bar1", "bar2", "xyz") == [True, True, False]

    def test_sunion(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support sunion")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2")
        cache.sadd("foo2", "bar2", "bar3")
        assert cache.sunion("foo1", "foo2") == {"bar1", "bar2", "bar3"}

    def test_sunionstore(self, cache: ValkeyCache):
        if isinstance(cache.client, ShardClient):
            pytest.skip("ShardClient doesn't support sunionstore")

        if isinstance(cache.client, DefaultClusterClient):
            pytest.skip("cluster client has a specific test")

        cache.sadd("foo1", "bar1", "bar2")
        cache.sadd("foo2", "bar2", "bar3")
        assert cache.sunionstore("foo3", "foo1", "foo2") == 3
        assert cache.smembers("foo3") == {"bar1", "bar2", "bar3"}

    def test_make_key(self, cache: ValkeyCache):
        assert cache.make_key("key", version=1, prefix="prefix") == "prefix:1:key"

    def test_make_pattern(self, cache: ValkeyCache):
        assert (
            cache.make_pattern("key_*", version=1, prefix="prefix") == "prefix:1:key_*"
        )



from collections.abc import Iterable

import pytest
from pytest_django.fixtures import SettingsWrapper
from pytest_mock import MockerFixture

from django.core.cache import DEFAULT_CACHE_ALIAS

from django_valkey.async_cache.cache import AsyncValkeyCache
from django_valkey.async_cache.client import AsyncDefaultClient

pytestmark = pytest.mark.anyio


@pytest.fixture
async def cache_client(cache: AsyncValkeyCache) -> Iterable[AsyncDefaultClient]:
    client = cache.client
    await client.aset("TestClientClose", 0)
    yield client
    await client.adelete("TestClientClose")


class TestClientClose:
    async def test_close_client_disconnect_default(
        self, cache_client: AsyncDefaultClient, mocker: MockerFixture
    ):
        mock = mocker.patch.object(
            cache_client.connection_factory, "disconnect", new_callable=mocker.AsyncMock
        )

        await cache_client.aclose()
        assert not mock.called

    async def test_close_disconnect_settings(
        self,
        cache_client: AsyncDefaultClient,
        settings: SettingsWrapper,
        mocker: MockerFixture,
    ):
        mock = mocker.patch.object(
            cache_client.connection_factory, "disconnect", new_callable=mocker.AsyncMock
        )

        settings.DJANGO_VALKEY_CLOSE_CONNECTION = True

        await cache_client.aclose()
        assert mock.called

    async def test_close_disconnect_settings_cache(
        self,
        cache_client: AsyncDefaultClient,
        mocker: MockerFixture,
        settings: SettingsWrapper,
    ):
        mock = mocker.patch.object(
            cache_client.connection_factory, "disconnect", new_callable=mocker.AsyncMock
        )

        settings.CACHES[DEFAULT_CACHE_ALIAS]["OPTIONS"]["CLOSE_CONNECTION"] = True
        await cache_client.aset("TestClientClose", 0)

        await cache_client.aclose()
        assert mock.called

    async def test_close_disconnect_client_options(
        self, cache_client: AsyncDefaultClient, mocker: MockerFixture
    ):
        mock = mocker.patch.object(
            cache_client.connection_factory, "disconnect", new_callable=mocker.AsyncMock
        )

        cache_client._options["CLOSE_CONNECTION"] = True

        await cache_client.aclose()
        assert mock.called



import pytest

from django.core.exceptions import ImproperlyConfigured

from django_valkey import pool


def test_connection_factory_redefine_from_opts():
    cf = pool.get_connection_factory(
        path="django_valkey.pool.ConnectionFactory",
        options={
            "CONNECTION_FACTORY": "django_valkey.pool.SentinelConnectionFactory",
            "SENTINELS": [("127.0.0.1", "26739")],
        },
    )
    assert cf.__class__.__name__ == "SentinelConnectionFactory"


@pytest.mark.parametrize(
    "conn_factory,expected",
    [
        (
            "django_valkey.pool.SentinelConnectionFactory",
            pool.SentinelConnectionFactory,
        ),
        ("django_valkey.pool.ConnectionFactory", pool.ConnectionFactory),
    ],
)
def test_connection_factory_opts(conn_factory: str, expected):
    cf = pool.get_connection_factory(
        path=None,
        options={
            "CONNECTION_FACTORY": conn_factory,
            "SENTINELS": [("127.0.0.1", "26739")],
        },
    )
    assert isinstance(cf, expected)


@pytest.mark.parametrize(
    "conn_factory,expected",
    [
        (
            "django_valkey.pool.SentinelConnectionFactory",
            pool.SentinelConnectionFactory,
        ),
        ("django_valkey.pool.ConnectionFactory", pool.ConnectionFactory),
    ],
)
def test_connection_factory_path(conn_factory: str, expected):
    cf = pool.get_connection_factory(
        path=conn_factory,
        options={
            "SENTINELS": [("127.0.0.1", "26739")],
        },
    )
    assert isinstance(cf, expected)


def test_connection_factory_no_sentinels():
    with pytest.raises(ImproperlyConfigured):
        pool.get_connection_factory(
            path=None,
            options={
                "CONNECTION_FACTORY": "django_valkey.pool.SentinelConnectionFactory",
            },
        )



import pytest


@pytest.fixture(scope="session")
def anyio_backend():
    return "asyncio"


# this keeps the event loop open for the entire test suite
@pytest.fixture(scope="session", autouse=True)
async def keepalive(anyio_backend):
    pass



from typing import Any

import msgspec

from django_valkey.serializers.base import BaseSerializer


class MsgSpecJsonSerializer(BaseSerializer):
    def dumps(self, value: Any) -> bytes:
        return msgspec.json.encode(value)

    def loads(self, value: bytes) -> Any:
        return msgspec.json.decode(value)


class MsgSpecMsgPackSerializer(BaseSerializer):
    def dumps(self, value: Any) -> bytes:
        return msgspec.msgpack.encode(value)

    def loads(self, value: bytes) -> Any:
        return msgspec.msgpack.decode(value)
